//
// gANTlab module for plane wave incidence
// © 10.11.2007 Ingo D. Rullhusen
//


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "module_header.h"

#include "module_data.h"
#include "rotmat.h"

using namespace std;



int MainData_version      = MainData_version_;
int FIELD_version         = FIELD_version_;
int module_header_version = module_header_version_;



int version = 1;
char *name = "Plane Wave Module";



extern "C" void* new_data(){ return( (void*) new ModuleData ); }

extern "C" void* dpl_data( void *di_ ){ return( (void*) new ModuleData( (ModuleData*)di_ ) ); }

extern "C" void del_data( void *di_ ){ delete (ModuleData*)di_; }



static double HYPOT( double x, double y, double z ){ return(sqrt(x*x+y*y+z*z)); }

const int matrix_para_no = 5;
const char *matrix_para[ matrix_para_no ] =
{
   "theta / °",
   "phi / °",
   "psi / °",
   "abs(E) / V/m",
   "arg(E) / °"
};



extern "C" void RHS_get_parameter( const MainData *MD, void *data_,
                                   string **param, string **value,
                                   int *no )
{
   ModuleData *data = (ModuleData*)data_;
   char sb[50];

   *no = matrix_para_no;
   *param = new string[*no];
   *value = new string[*no];

   int id = 0;

   sprintf(sb,"%g",data->theta * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->phi * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->psi * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;


   sprintf(sb,"%g", abs(data->E0) );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", arg(data->E0) * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];  
   (*value)[id] = sb;
   id++;

}


extern "C" void RHS_set_parameter( const MainData *MD, void *data_,
                                   const string *param, const string *value,
                                   int no )
{
   ModuleData *data = (ModuleData*)data_;
   int j;
   double absE, argE;

   absE = abs(data->E0);
   argE = arg(data->E0);

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<matrix_para_no; j++ ) if ( matrix_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { data->theta = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 1: { data->phi   = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 2: { data->psi   = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 3: { absE = atof( value[i].c_str() ); break; }
         case 4: { argE = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         default: throw( "unknown parameter" );
      }
   }

   data->E0 = complex<double>( absE * cos(argE), absE * sin(argE) );

   double rot[3][3];
   rotmat( rot, 1.0, data->psi, 0.0, 0.0, data->theta, data->phi );
   data->uE[0] = rot[0][0];
   data->uE[1] = rot[1][0];
   data->uE[2] = rot[2][0];

   data->uH[0] = rot[0][1];
   data->uH[1] = rot[1][1];
   data->uH[2] = rot[2][1];

   data->uk[0] = rot[0][2];
   data->uk[1] = rot[1][2];
   data->uk[2] = rot[2][2];
}


#ifdef HAVE_LIBGL
extern "C" void RHS_add_to_OGLlist( const MainData *MD, void *data_, GLuint gl_list )
{
   const GLfloat clr_E[4] = { 0.90, 0.00, 0.00, 1.0 };
   const GLfloat clr_H[4] = { 0.25, 0.25, 1.00, 1.0 };
   const GLfloat clr_k[4] = { 0.90, 0.90, 0.00, 1.0 };
   ModuleData *data = (ModuleData*)data_;

   double R  = 0.5*HYPOT( MD->size_x, MD->size_y, MD->size_z );

// E
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_E );
   glBegin(GL_LINES);

   glVertex3d( -0.10*R*data->uE[0], -0.10*R*data->uE[1], -0.10*R*data->uE[2] );
   glVertex3d(  0.50*R*data->uE[0],  0.50*R*data->uE[1],  0.50*R*data->uE[2] );

   glEnd();

// H
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_H );
   glBegin(GL_LINES);

   glVertex3d( -0.10*R*data->uH[0], -0.10*R*data->uH[1], -0.10*R*data->uH[2] );
   glVertex3d(  0.50*R*data->uH[0],  0.50*R*data->uH[1],  0.50*R*data->uH[2] );

   glEnd();

// k
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_k );
   glBegin(GL_LINES);

   glVertex3d( -0.10*R*data->uk[0], -0.10*R*data->uk[1], -0.10*R*data->uk[2] );
   glVertex3d(  0.90*R*data->uk[0],  0.90*R*data->uk[1],  0.90*R*data->uk[2] );

   glEnd();
}
#endif


extern "C" void RHS_Einc( const MainData *MD, void *data_, double x, double y, double z,
                          complex<double> E[3] )
{
   ModuleData *data = (ModuleData*)data_;

   complex<double> expo = exp( complex<double>( 0.0, -MD->k0*( data->uk[0]*x+data->uk[1]*y+data->uk[2]*z ) ) );

   expo *= data->E0;

   E[0] = data->uE[0] * expo;
   E[1] = data->uE[1] * expo;
   E[2] = data->uE[2] * expo;
}
