//
// gANTlab module for RWG basis functions
// (c) 10.11.2007 Ingo D. Rullhusen
//

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <complex>
#include <fstream>

#include <iostream>

#include "module_header.h"

#include "module_data.h"
#include "triquad.h"
#include "green.h"

using namespace std;



int MainData_version      = MainData_version_;
int FIELD_version         = FIELD_version_;
int module_header_version = module_header_version_;



int version = 1;
char *name = "RWG Module";



extern "C" void* new_data(){ return( (void*) new ModuleData ); }

extern "C" void* dpl_data( void *di_ ){ return( (void*) new ModuleData( (ModuleData*)di_ ) ); }

extern "C" void del_data( void *di_ ){ delete (ModuleData*)di_; }


static double HYPOT( double x, double y, double z ){ return(sqrt(x*x+y*y+z*z)); }


const int matrix_para_no = 8;
const char *matrix_para[ matrix_para_no ] =
{
   "OFF file",
   "mesh unit / m",
   "center point x / m",
   "center point y / m",
   "center point z / m",
   "theta / °",
   "phi / °",
   "psi / °"
};

extern "C" void matrix_get_parameter( const MainData *MD, void *data_,
                                      string **param, string **value, int *no )
{
   ModuleData *data = (ModuleData*)data_;
   char sb[50];

   *no = matrix_para_no;
   *param = new string[*no];
   *value = new string[*no];

   int id = 0;

   (*param)[id] = matrix_para[id];
   (*value)[id] = data->off_file_name;
   id++;

   sprintf(sb,"%g",data->mesh_unit);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cp_x);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cp_y);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->cp_z);
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->theta * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->phi * 180.0 / M_PI );
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g", data->psi * 180.0 / M_PI );  
   (*param)[id] = matrix_para[id];
   (*value)[id] = sb;
   id++;

}

extern "C" void matrix_set_parameter( const MainData *MD, void *data_,
                                      const string *param, const string *value, int no )
{
   ModuleData *data = (ModuleData*)data_;
   int j;

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<matrix_para_no; j++ ) if ( matrix_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { data->off_file_name = value[i]; break; }
         case 1: { data->mesh_unit = atof( value[i].c_str() ); break; }
         case 2: { data->cp_x = atof( value[i].c_str() ); break; }
         case 3: { data->cp_y = atof( value[i].c_str() ); break; }
         case 4: { data->cp_z = atof( value[i].c_str() ); break; }
         case 5: { data->theta = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 6: { data->phi   = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         case 7: { data->psi   = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         default: throw( "unknown parameter" );
      }
   }

}


extern "C" int matrix_geo_setup( const MainData *MD, void *data_,
                                 string **omsg, int *nomsg )
{
   ModuleData *data = (ModuleData*)data_;

   data->read_off_file();
   delete data->grf;
   data->grf = new GREEN( MD->k0 );

   *nomsg = 3;
   *omsg = new string[*nomsg];

   const int buffsize = 64;
   char buff[buffsize];
   snprintf( buff, buffsize, "no. of points   : %u", (unsigned)data->plist.size() );
   (*omsg)[0] = buff;
   snprintf( buff, buffsize, "no. of triangles: %u", (unsigned)data->tlist.size() );
   (*omsg)[1] = buff;
   snprintf( buff, buffsize, "no. of RWG edges: %u", (unsigned)data->velist.size() );
   (*omsg)[2] = buff;

   return( data->velist.size() );
}


extern "C" void matrix_renumber( const MainData *MD, void *data_, int *perm, int *rperm )
{
   ModuleData *data = (ModuleData*)data_;

   for ( TRIANGLE_LIST::iterator it = data->tlist.begin();
         it != data->tlist.end();
         it++ )
   {
      if ( it->ve1 > -1 ) it->ve1 = perm[ it->ve1 ];
      if ( it->ve2 > -1 ) it->ve2 = perm[ it->ve2 ];
      if ( it->ve3 > -1 ) it->ve3 = perm[ it->ve3 ];
   }

   vEDGE_LIST vet;
   vet.clear();

   for ( unsigned int i=0; i<data->velist.size(); i++ )
   {
      vet.push_back( data->velist[ rperm[i] ] );
   }

   data->velist.swap(vet);
   vet.clear();
}


extern "C" void matrix_bounding_box( const MainData *MD, void *data_,
                                     double *cpos_x, double *cpos_y, double *cpos_z,
                                     double *size_x, double *size_y, double *size_z )
{
   ModuleData *data = (ModuleData*)data_;

   if ( data->plist.size() < 1 )
   {
      *cpos_x = MD->cpos_x;
      *cpos_y = MD->cpos_y;
      *cpos_z = MD->cpos_z;

      *size_x = *size_y = *size_z = 0.0;

      return;
   }

   double minx, maxx, miny, maxy, minz, maxz;
   minx = maxx = data->plist[0].x;
   miny = maxy = data->plist[0].y;
   minz = maxz = data->plist[0].z;

   for ( POINT_LIST::iterator it = data->plist.begin()+1;
         it != data->plist.end();
         it++ )
   {
      minx = ( minx < it->x ? minx : it->x );
      miny = ( miny < it->y ? miny : it->y );
      minz = ( minz < it->z ? minz : it->z );

      maxx = ( maxx > it->x ? maxx : it->x );
      maxy = ( maxy > it->y ? maxy : it->y );
      maxz = ( maxz > it->z ? maxz : it->z );
   }

   *size_x = maxx - minx;
   *size_y = maxy - miny;
   *size_z = maxz - minz;

   *cpos_x = 0.5 * ( maxx + minx );
   *cpos_y = 0.5 * ( maxy + miny );
   *cpos_z = 0.5 * ( maxz + minz );
}


#ifdef HAVE_LIBGL
extern "C" void matrix_add_to_OGLlist( const MainData *MD, void *data_, GLuint list )
{
   ModuleData *data = (ModuleData*)data_;

   data->ogl_triangles( list );
}
#endif


extern "C"  void matrix_bfkt_bounding_box( const MainData *MD, void *data_, int nbfkt,
                                           int *sno,
                                           double **cpos_x, double **cpos_y, double **cpos_z,
                                           double **size_x, double **size_y, double **size_z )
{
   ModuleData *data = (ModuleData*)data_;

   if ( nbfkt < 0 || nbfkt > (int)data->velist.size() )
      throw( "invalid basis function number" );

   double minx, maxx, miny, maxy, minz, maxz;
   POINT *P;

   *sno = 1;

   *cpos_x = new double[*sno];
   *cpos_y = new double[*sno];
   *cpos_z = new double[*sno];

   *size_x = new double[*sno];
   *size_y = new double[*sno];
   *size_z = new double[*sno];

   P = &data->plist[ data->velist[nbfkt].p1 ];
   minx = maxx = P->x;
   miny = maxy = P->y;
   minz = maxz = P->z;

   P = &data->plist[ data->velist[nbfkt].p2 ];
   minx = ( minx < P->x ? minx : P->x );
   miny = ( miny < P->y ? miny : P->y );
   minz = ( minz < P->z ? minz : P->z );

   maxx = ( maxx > P->x ? maxx : P->x );
   maxy = ( maxy > P->y ? maxy : P->y );
   maxz = ( maxz > P->z ? maxz : P->z );

   P = &data->plist[ data->velist[nbfkt].spp ];
   minx = ( minx < P->x ? minx : P->x );
   miny = ( miny < P->y ? miny : P->y );
   minz = ( minz < P->z ? minz : P->z );

   maxx = ( maxx > P->x ? maxx : P->x );
   maxy = ( maxy > P->y ? maxy : P->y );
   maxz = ( maxz > P->z ? maxz : P->z );

   P = &data->plist[ data->velist[nbfkt].spm ];
   minx = ( minx < P->x ? minx : P->x );
   miny = ( miny < P->y ? miny : P->y );
   minz = ( minz < P->z ? minz : P->z );

   maxx = ( maxx > P->x ? maxx : P->x );
   maxy = ( maxy > P->y ? maxy : P->y );
   maxz = ( maxz > P->z ? maxz : P->z );


   (*size_x)[0] = maxx - minx;
   (*size_y)[0] = maxy - miny;
   (*size_z)[0] = maxz - minz;

   (*cpos_x)[0] = 0.5 * ( maxx + minx );
   (*cpos_y)[0] = 0.5 * ( maxy + miny );
   (*cpos_z)[0] = 0.5 * ( maxz + minz );
}


extern "C" void matrix_bfkt_moment( const MainData *MD, void *data_,
                                    int veno, int sno,
                                    double cpos_x, double cpos_y, double cpos_z,
                                    double len_x, double len_y, double len_z,
                                    int nx, int ny, int nz,
                                    double *mx, double *my, double *mz, double *md )
{
   ModuleData *data = (ModuleData*)data_;
   triquad irule(7);
   int t;
   double ipx, ipy, ipz, dx, dy, dz, powpow;

   *mx = *my = *mz = *md = 0.0;

// + triangle
   t = data->velist[veno].trip;
   for ( int i=0; i<irule.no; i++ )
   {

      ipx = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].x
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].x
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].x;

      ipy = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].y
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].y
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].y;

      ipz = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].z
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].z
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].z;


      dx = ( ipx - cpos_x ) / len_x;
      dy = ( ipy - cpos_y ) / len_y;
      dz = ( ipz - cpos_z ) / len_z;

      powpow =  pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz );

// RWG functions
      *mx += irule.weights[i] * 0.5 * powpow
             * ( ipx - data->plist[ data->velist[veno].spp ].x );
      *my += irule.weights[i] * 0.5 * powpow
             * ( ipy - data->plist[ data->velist[veno].spp ].y );
      *mz += irule.weights[i] * 0.5 * powpow
             * ( ipz - data->plist[ data->velist[veno].spp ].z );

      *md += irule.weights[i] * powpow;
   }

// - triangle
   t = data->velist[veno].trim;
   for ( int i=0; i<irule.no; i++ )
   {

      ipx = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].x
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].x
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].x;

      ipy = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].y
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].y
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].y;

      ipz = irule.lambda1[i] * data->plist[ data->tlist[t].p1 ].z
            + irule.lambda2[i] * data->plist[ data->tlist[t].p2 ].z
            + irule.lambda3[i] * data->plist[ data->tlist[t].p3 ].z;


      dx = ( ipx - cpos_x ) / len_x;
      dy = ( ipy - cpos_y ) / len_y;
      dz = ( ipz - cpos_z ) / len_z;

      powpow =  pow( dx, nx ) * pow( dy, ny ) * pow( dz, nz );

// RWG functions
      *mx -= irule.weights[i] * 0.5 * powpow
             * ( ipx - data->plist[ data->velist[veno].spm ].x );
      *my -= irule.weights[i] * 0.5 * powpow
             * ( ipy - data->plist[ data->velist[veno].spm ].y );
      *mz -= irule.weights[i] * 0.5 * powpow
             * ( ipz - data->plist[ data->velist[veno].spm ].z );

      *md -= irule.weights[i] * powpow;
   }

}


extern "C"  complex<double> matrix_ele( const MainData *MD, void *data_, int e1, int e2 )
{
   ModuleData *data = (ModuleData*)data_;

   const complex<double> cNULL(0.0,0.0);
   const complex<double> cONE(1.0,0.0);
   const complex<double> cJ(0.0,1.0);

   triquad oI(3), iI(7);

   bool singu;
   int atri, qtri;
   double rv1[3], rv2[3], area;
   complex<double> zval, A[3], phi,
                   Gd[3][3], Gs, v[3], J[3], div, J0[3], div0;
   POINT ap, qp;

#ifdef BOUNDARY_CHECK
   if( e1 < 0 || e1 >= (int)data->velist.size() ||
       e2 < 0 || e2 >= (int)data->velist.size()    )
   {
      cerr << "Error in RWGModule::matrix_ele(): line " << __LINE__ << "!" << endl;
      exit(1);
   }
#endif

   complex<double> jomega = cJ*MD->omega;

   zval = cNULL;

// triangle 1 p
   atri = data->velist[e1].trip;

   for ( int ip1=0; ip1<oI.no; ip1++ )
   {
      ap.x = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].x
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].x
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].x;
      ap.y = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].y
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].y
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].y;
      ap.z = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].z
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].z
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].z;


      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi = cNULL;

// triangle 2 p
      qtri = data->velist[e2].trip;
      data->triangle_center( qtri, qp );

      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( ap.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
            J0[1] = ( ap.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
            J0[2] = ( ap.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;
            div0 = cONE;
         }
         else
         {
            J0[0] = ( qp.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
            J0[1] = ( qp.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
            J0[2] = ( qp.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;
            div0 = cONE;
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( qp.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
         rv2[1] = ( qp.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
         rv2[2] = ( qp.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;

         div = cONE;

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = data->triangle_area(qtri);
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

// triangle 2 m
      qtri = data->velist[e2].trim;
      data->triangle_center( qtri, qp );

      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( data->plist[ data->velist[e2].spm ].x - ap.x ) * 0.5;
            J0[1] = ( data->plist[ data->velist[e2].spm ].y - ap.y ) * 0.5;
            J0[2] = ( data->plist[ data->velist[e2].spm ].z - ap.z ) * 0.5;
            div0 = -cONE;
         }
         else
         {
            J0[0] = ( data->plist[ data->velist[e2].spm ].x - qp.x ) * 0.5;
            J0[1] = ( data->plist[ data->velist[e2].spm ].y - qp.y ) * 0.5;
            J0[2] = ( data->plist[ data->velist[e2].spm ].z - qp.z ) * 0.5;
            div0 = -cONE;
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;

         rv2[0] = ( data->plist[ data->velist[e2].spm ].x - qp.x ) * 0.5;
         rv2[1] = ( data->plist[ data->velist[e2].spm ].y - qp.y ) * 0.5;
         rv2[2] = ( data->plist[ data->velist[e2].spm ].z - qp.z ) * 0.5;

         div = -cONE;

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = data->triangle_area(qtri);
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

      rv1[0] = ( ap.x - data->plist[ data->velist[e1].spp ].x ) * 0.5;
      rv1[1] = ( ap.y - data->plist[ data->velist[e1].spp ].y ) * 0.5;
      rv1[2] = ( ap.z - data->plist[ data->velist[e1].spp ].z ) * 0.5;

      zval += oI.weights[ip1] * (
                 - jomega * ( rv1[0]*A[0] + rv1[1]*A[1] + rv1[2]*A[2] )
                 - 1.0 * phi / jomega
                                 );

   }

// triangle 1 m
   atri = data->velist[e1].trim;

   for ( int ip1=0; ip1<oI.no; ip1++ )
   {
      ap.x = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].x
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].x
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].x;
      ap.y = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].y
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].y
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].y;
      ap.z = oI.lambda1[ip1] * data->plist[ data->tlist[atri].p1 ].z
             + oI.lambda2[ip1] * data->plist[ data->tlist[atri].p2 ].z
             + oI.lambda3[ip1] * data->plist[ data->tlist[atri].p3 ].z;


      A[0] = cNULL;
      A[1] = cNULL;
      A[2] = cNULL;
      phi = cNULL;

// triangle 2 p
      qtri = data->velist[e2].trip;
      data->triangle_center( qtri, qp );

      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( ap.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
            J0[1] = ( ap.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
            J0[2] = ( ap.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;
            div0 = cONE;
         }
         else
         {
            J0[0] = ( qp.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
            J0[1] = ( qp.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
            J0[2] = ( qp.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;
            div0 = cONE;
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( qp.x - data->plist[ data->velist[e2].spp ].x ) * 0.5;
         rv2[1] = ( qp.y - data->plist[ data->velist[e2].spp ].y ) * 0.5;
         rv2[2] = ( qp.z - data->plist[ data->velist[e2].spp ].z ) * 0.5;

         div = cONE;

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = data->triangle_area(qtri);
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

// triangle 2 m
      qtri = data->velist[e2].trim;
      data->triangle_center( qtri, qp );

      if ( data->triangle_neigh(atri,qtri) )
      {
         singu = true;
         if ( atri == qtri )
         {
            J0[0] = ( data->plist[ data->velist[e2].spm ].x - ap.x ) * 0.5;
            J0[1] = ( data->plist[ data->velist[e2].spm ].y - ap.y ) * 0.5;
            J0[2] = ( data->plist[ data->velist[e2].spm ].z - ap.z ) * 0.5;

            div0 = -cONE;
         }
         else
         {
            J0[0] = ( data->plist[ data->velist[e2].spm ].x - qp.x ) * 0.5;
            J0[1] = ( data->plist[ data->velist[e2].spm ].y - qp.y ) * 0.5;
            J0[2] = ( data->plist[ data->velist[e2].spm ].z - qp.z ) * 0.5;

            div0 = -cONE;
         }
      }
      else
         singu = false;

      for ( int ip2=0; ip2<iI.no; ip2++ )
      {
         qp.x = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].x
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].x
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].x;
         qp.y = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].y
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].y
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].y;
         qp.z = iI.lambda1[ip2] * data->plist[ data->tlist[qtri].p1 ].z
                + iI.lambda2[ip2] * data->plist[ data->tlist[qtri].p2 ].z
                + iI.lambda3[ip2] * data->plist[ data->tlist[qtri].p3 ].z;


         rv2[0] = ( data->plist[ data->velist[e2].spm ].x - qp.x ) * 0.5;
         rv2[1] = ( data->plist[ data->velist[e2].spm ].y - qp.y ) * 0.5;
         rv2[2] = ( data->plist[ data->velist[e2].spm ].z - qp.z ) * 0.5;

         div = -cONE;

         if ( singu )
         {
            J[0] = rv2[0];
            J[1] = rv2[1];
            J[2] = rv2[2];
            data->grf->EP_green_ex( v, Gs, J, div, J0, div0, ap, qp );

            A[0] += iI.weights[ip2] * v[0];  // /triangle
            A[1] += iI.weights[ip2] * v[1];  // /triangle
            A[2] += iI.weights[ip2] * v[2];  // /triangle

            phi += iI.weights[ip2] * Gs;  // /triangle
         }
         else
         {
            data->grf->EP_green( Gd, Gs, ap, qp );

            A[0] += iI.weights[ip2] *
                    ( Gd[0][0]*rv2[0] + Gd[0][1]*rv2[1] + Gd[0][2]*rv2[2] );  // /triangle
            A[1] += iI.weights[ip2] *
                    ( Gd[1][0]*rv2[0] + Gd[1][1]*rv2[1] + Gd[1][2]*rv2[2] );  // /triangle
            A[2] += iI.weights[ip2] *
                    ( Gd[2][0]*rv2[0] + Gd[2][1]*rv2[1] + Gd[2][2]*rv2[2] );  // /triangle

            phi += iI.weights[ip2] * Gs * div;  // /triangle
         }
      }

      if ( singu )
      {
         area = data->triangle_area(qtri);
         J0[0] /= area;
         J0[1] /= area;
         J0[2] /= area;
         div0  /= area;

         data->grf->EP_green_re( v, div, J0, div0,
                                 data->plist[ data->tlist[qtri].p1 ],
                                    data->plist[ data->tlist[qtri].p2 ],
                                       data->plist[ data->tlist[qtri].p3 ], ap );
         A[0] += v[0];
         A[1] += v[1];
         A[2] += v[2];
         phi += div;
      }

      rv1[0] = ( data->plist[ data->velist[e1].spm ].x - ap.x ) * 0.5;
      rv1[1] = ( data->plist[ data->velist[e1].spm ].y - ap.y ) * 0.5;
      rv1[2] = ( data->plist[ data->velist[e1].spm ].z - ap.z ) * 0.5;

      zval += oI.weights[ip1] * (
                 - jomega * ( rv1[0]*A[0] + rv1[1]*A[1] + rv1[2]*A[2] )
                 - (-1.0) * phi / jomega
                                 );

   }

   return(zval);
}





//extern "C" void RHS_get_parameter( const MainData *MD, void *data_,
//                                   string **param, string **value, int *no )

//extern "C" void RHS_set_parameter( const MainData *MD, void *data_,
//                                   const string *param, const string *value, int no )

//extern "C" void RHS_add_to_OGLlist( const MainData *MD, void *data_, GLuint list ) {}





const int extract_para_no = 9;
const char *extract_para[ extract_para_no ] =
{ 
   "J probe file",
   "start point x / m",
   "start point y / m",  
   "start point z / m",
   "no. of samples",
   "end point x / m",
   "end point y / m",
   "end point z / m",
   "psi / °"
};

extern "C" void extract_get_parameter( const MainData *MD, void *data_,
                                       string **param, string **value,
                                       int *no )
{
   ModuleData *data = (ModuleData*)data_;
   char sb[50];

   *no = extract_para_no;
   *param = new string[*no];
   *value = new string[*no];   

   int id = 0;

   (*param)[id] = extract_para[id];
   (*value)[id] = data->J_probe_file_name;
   id++;


   sprintf(sb,"%g",data->startpkt[0]);
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->startpkt[1]);
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->startpkt[2]);
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;


   sprintf(sb,"%i",data->noofpkt);
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;


   sprintf(sb,"%g",data->endpkt[0]);
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->endpkt[1]);
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;

   sprintf(sb,"%g",data->endpkt[2]);
   (*param)[id] = extract_para[id];
   (*value)[id] = sb;
   id++;


   sprintf(sb,"%g", data->extract_psi * 180.0 / M_PI );
   (*param)[id] = extract_para[id]; 
   (*value)[id] = sb;
   id++;

}

extern "C" void extract_set_parameter( const MainData *MD, void *data_,
                                       const string *param, const string *value,
                                       int no )
{
   ModuleData *data = (ModuleData*)data_;
   int j;

   for ( int i=0; i<no; i++ )
   {
      for ( j=0; j<extract_para_no; j++ ) if ( extract_para[j] == param[i] ) break;

      switch(j)
      {
         case 0: { data->J_probe_file_name = value[i]; break; }
         case 1: { data->startpkt[0] = atof( value[i].c_str() ); break; }
         case 2: { data->startpkt[1] = atof( value[i].c_str() ); break; }
         case 3: { data->startpkt[2] = atof( value[i].c_str() ); break; }
         case 4: { data->noofpkt = atoi( value[i].c_str() ); break; }
         case 5: { data->endpkt[0] = atof( value[i].c_str() ); break; }
         case 6: { data->endpkt[1] = atof( value[i].c_str() ); break; }
         case 7: { data->endpkt[2] = atof( value[i].c_str() ); break; }
         case 8: { data->extract_psi = atof( value[i].c_str() ) * M_PI / 180.0; break; }
         default: throw( "unknown parameter" );
      }
   }

   data->setup_exvec();
}


#ifdef HAVE_LIBGL
extern "C" void extract_add_to_OGLlist( const MainData *MD, void *data_, GLuint list )
{
   const GLfloat clr_line[4]   = { 1.00, 0.30, 0.30, 1.0 };
   const GLfloat clr_cross1[4] = { 0.30, 1.00, 0.30, 1.0 };
   const GLfloat clr_cross2[4] = { 0.40, 0.40, 1.00, 1.0 };
   ModuleData *data = (ModuleData*)data_;

// line
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_line );

   glBegin(GL_LINES);

   glVertex3d( data->startpkt[0], data->startpkt[1], data->startpkt[2] );
   glVertex3d( data->endpkt[0], data->endpkt[1], data->endpkt[2] );

   double R = HYPOT( data->endpkt[0]-data->startpkt[0],
                     data->endpkt[1]-data->startpkt[1],
                     data->endpkt[2]-data->startpkt[2]  );
/***
   glVertex3d( -0.02*R*data->lv[0] + data->startpkt[0],
               -0.02*R*data->lv[1] + data->startpkt[1],
               -0.02*R*data->lv[2] + data->startpkt[2]  );
   glVertex3d( 0.1*R*data->lv[0] + data->startpkt[0],
               0.1*R*data->lv[1] + data->startpkt[1],
               0.1*R*data->lv[2] + data->startpkt[2]  );
***/

   glEnd();

// cross1
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_cross1 );

   glBegin(GL_LINES);

   glVertex3d( -0.02*R*data->c1v[0] + data->startpkt[0],
               -0.02*R*data->c1v[1] + data->startpkt[1],
               -0.02*R*data->c1v[2] + data->startpkt[2]  );
   glVertex3d( 0.1*R*data->c1v[0] + data->startpkt[0],
               0.1*R*data->c1v[1] + data->startpkt[1],
               0.1*R*data->c1v[2] + data->startpkt[2]  );

   glEnd();

// cross2
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_cross2 );

   glBegin(GL_LINES);

   glVertex3d( -0.02*R*data->c2v[0] + data->startpkt[0],
               -0.02*R*data->c2v[1] + data->startpkt[1],
               -0.02*R*data->c2v[2] + data->startpkt[2]  );
   glVertex3d( 0.1*R*data->c2v[0] + data->startpkt[0],
               0.1*R*data->c2v[1] + data->startpkt[1],
               0.1*R*data->c2v[2] + data->startpkt[2]  );

   glEnd();
}
#endif


extern "C" void extract_run( const MainData *MD, void *data_, bool dummy,
                             complex<double> *JV, int no, const FIELD *f,
                             string **omsg, int *nomsg )
{
   const complex<double> cNULL(0.0,0.0);
   ModuleData *data = (ModuleData*)data_;
   fstream jfile;

   string name = MD->PrjPath + "/" + data->J_probe_file_name;

   jfile.open( name.c_str(), ios::out );
   if ( ! jfile.is_open() ) throw( "cannot open output file" );

   jfile << "# pos & Re Jref & Im Jref & Re Jcross1 & Im Jcross1 & Re Jcross2 & Im Jcross2" << endl;

   int trino;
   POINT pkt;
   double area, nx, ny, nz, r;
   double DD = ( data->noofpkt < 2
                 ? 0.0
                 : HYPOT(  data->endpkt[0]-data->startpkt[0],
                           data->endpkt[1]-data->startpkt[1],
                           data->endpkt[2]-data->startpkt[2]  )
                   / ( data->noofpkt - 1 )
               );
   complex<double> Jx, Jy, Jz, J;
   TRIANGLE *tri;

   for ( int p=0; p<data->noofpkt; p++ )
   {
      Jx = Jy = Jz = cNULL;

      pkt.x = data->startpkt[0] + p*DD*data->lv[0];
      pkt.y = data->startpkt[1] + p*DD*data->lv[1];
      pkt.z = data->startpkt[2] + p*DD*data->lv[2];

      trino = data->near_triangle( pkt );
      area = data->triangle_normal( trino, nx, ny, nz );

      r = pkt.x*nx + pkt.y*ny + pkt.z*nz;
      pkt.x -= r*nx;
      pkt.y -= r*ny;
      pkt.z -= r*nz;

      tri = &( data->tlist[trino] );
// edge 1
      if ( tri->sign.sp1 != 0 )
      {
         Jx += JV[ tri->ve1 ] * (double)tri->sign.sp1
               * ( pkt.x - data->plist[ tri->p1 ].x ) * 0.5 / area;
         Jy += JV[ tri->ve1 ] * (double)tri->sign.sp1
               * ( pkt.y - data->plist[ tri->p1 ].y ) * 0.5 / area;
         Jz += JV[ tri->ve1 ] * (double)tri->sign.sp1
               * ( pkt.z - data->plist[ tri->p1 ].z ) * 0.5 / area;
      }
// edge 2
      if ( tri->sign.sp2 != 0 )
      {
         Jx += JV[ tri->ve2 ] * (double)tri->sign.sp2
               * ( pkt.x - data->plist[ tri->p2 ].x ) * 0.5 / area;
         Jy += JV[ tri->ve2 ] * (double)tri->sign.sp2
               * ( pkt.y - data->plist[ tri->p2 ].y ) * 0.5 / area;
         Jz += JV[ tri->ve2 ] * (double)tri->sign.sp2
               * ( pkt.z - data->plist[ tri->p2 ].z ) * 0.5 / area;
      }
// edge 3
      if ( tri->sign.sp3 != 0 )
      {
         Jx += JV[ tri->ve3 ] * (double)tri->sign.sp3
               * ( pkt.x - data->plist[ tri->p3 ].x ) * 0.5 / area;
         Jy += JV[ tri->ve3 ] * (double)tri->sign.sp3
               * ( pkt.y - data->plist[ tri->p3 ].y ) * 0.5 / area;
         Jz += JV[ tri->ve3 ] * (double)tri->sign.sp3
               * ( pkt.z - data->plist[ tri->p3 ].z ) * 0.5 / area;
      }

      jfile << p*DD << "  ";
      J = data->lv[0]*Jx + data->lv[1]*Jy + data->lv[2]*Jz;
      jfile << J.real() << " " << J.imag() << "  ";
      J = data->c1v[0]*Jx + data->c1v[1]*Jy + data->c1v[2]*Jz;
      jfile << J.real() << " " << J.imag() << "  ";
      J = data->c2v[0]*Jx + data->c2v[1]*Jy + data->c2v[2]*Jz;
      jfile << J.real() << " " << J.imag() << endl;
   }

   jfile.close();

   *nomsg = 0;
   *omsg = NULL;
}

