//
// Class for greens function
// (c) 2.11.2006 Ingo D. Rullhusen
//

#include <float.h>

#include "natconst.h"

#include "green.h"
#include "singularity.h"


static const complex<double> cNULL(0.0,0.0);
static const complex<double> cONE(1.0,0.0);
static const complex<double> cJ(0.0,1.0);

static double SQU( double x ){ return(x*x); }


GREEN::GREEN( double k0_ ) : k0(k0_)
{
   prefac_v = MY0/(4.0*M_PI);
   prefac_div = 1.0/(4.0*M_PI*EPS0);
   prefac_E = ZF0/(4.0*M_PI);
   prefac_ff = cJ*k0*ZF0/(4.0*M_PI);
}


void GREEN::EP_green( complex<double> grfv[3][3], complex<double> &grfdiv,
                      const POINT ap, const POINT qp )
{
   double R;
   complex<double> expo;

   R = sqrt( SQU(qp.x - ap.x) + SQU(qp.y - ap.y) + SQU(qp.z - ap.z) );
   expo = exp( -cJ*k0*R ) / R;

   grfv[0][0] = grfv[1][1] = grfv[2][2] = prefac_v * expo;
   grfv[1][0] = grfv[0][1] = grfv[2][0] = grfv[0][2] = grfv[2][1] = grfv[1][2] = cNULL;

   grfdiv = prefac_div * expo;
}

void GREEN::EP_green_ex( complex<double> pv[3], complex<double> &gdiv,
                         const complex<double> J[3], complex<double> div,       
                         const complex<double> J0[3], complex<double> div0,
                         const POINT ap, const POINT qp )
{
   double R;
   complex<double> expo;

   R = sqrt( SQU(ap.x-qp.x) + SQU(ap.y-qp.y) + SQU(ap.z-qp.z) );
   if ( R > DBL_EPSILON )
   {
      expo = exp( -cJ*k0*R );

      pv[0] = prefac_v * ( expo*J[0] - J0[0] ) / R;
      pv[1] = prefac_v * ( expo*J[1] - J0[1] ) / R;
      pv[2] = prefac_v * ( expo*J[2] - J0[2] ) / R;

      gdiv = prefac_div * ( expo*div - div0 ) / R;
   }
   else
   {
      expo = complex<double>(0.0,-k0);

      pv[0] = prefac_v * expo * J0[0];
      pv[1] = prefac_v * expo * J0[1];
      pv[2] = prefac_v * expo * J0[2];

      gdiv = prefac_div * expo * div0;
   }
}

void GREEN::EP_green_re( complex<double> pv[3], complex<double> &div, 
                         const complex<double> J0[3], complex<double> div0,
                         const POINT tp1,
                            const POINT tp2,
                               const POINT tp3,
                         const POINT ap )
{
   const int pno = 4;
   double px[pno], py[pno], sgI, x[3], y[3], z[3], lx, ly_x, ly_y;

   x[0] = tp2.x - tp1.x;
   x[1] = tp2.y - tp1.y;
   x[2] = tp2.z - tp1.z;
   lx = sqrt( x[0]*x[0] + x[1]*x[1] + x[2]*x[2] );
   x[0] /= lx;
   x[1] /= lx;
   x[2] /= lx;

   y[0] = tp3.x - tp1.x;
   y[1] = tp3.y - tp1.y;
   y[2] = tp3.z - tp1.z;
   ly_x = x[0]*y[0] + x[1]*y[1] + x[2]*y[2];
   y[0] -= ly_x*x[0];
   y[1] -= ly_x*x[1];
   y[2] -= ly_x*x[2];
   ly_y = sqrt( y[0]*y[0] + y[1]*y[1] + y[2]*y[2] );
   y[0] /= ly_y;
   y[1] /= ly_y;
   y[2] /= ly_y;

   z[0] = x[1]*y[2] - x[2]*y[1];
   z[1] = x[2]*y[0] - x[0]*y[2];
   z[2] = x[0]*y[1] - x[1]*y[0];


   px[0] = 0.0;   py[0] = 0.0;
   px[1] = lx;    py[1] = 0.0;
   px[2] = ly_x;  py[2] = ly_y;
   px[3] = px[0]; py[3] = py[0];

   sgI = Psing( pno, px, py, 
                x[0]*(ap.x-tp1.x) + x[1]*(ap.y-tp1.y) + x[2]*(ap.z-tp1.z),
                y[0]*(ap.x-tp1.x) + y[1]*(ap.y-tp1.y) + y[2]*(ap.z-tp1.z),
                z[0]*(ap.x-tp1.x) + z[1]*(ap.y-tp1.y) + z[2]*(ap.z-tp1.z)  );

   pv[0] = prefac_v * sgI * J0[0];
   pv[1] = prefac_v * sgI * J0[1];
   pv[2] = prefac_v * sgI * J0[2];

   div = prefac_div * sgI * div0;
}
