//
// module data
// (c) 10.11.2007 Ingo D. Rullhusen
//

#include <complex>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>

#include <math.h>
#include <float.h>
#include <sys/stat.h>


#include "module_data.h"
#include "rotmat.h"


static const double DIST_EPSILON = 1e-10;

static double HYPOT( double x, double y, double z ){ return(sqrt(x*x+y*y+z*z)); }


ModuleData::ModuleData()
{
   off_file_name = "";
   mesh_unit = 1.0;

   cp_x = cp_y = cp_z = 0.0;
   theta = phi = psi = 0.0;

   grf = NULL;

   plist.clear();
   tlist.clear();
   velist.clear();

   J_probe_file_name = "";
   noofpkt = 0;
   startpkt[0] = startpkt[1] = startpkt[2] = 0.0;
   endpkt[0] = endpkt[1] = endpkt[2] = 0.0;
   extract_psi = 0.0;

   setup_exvec();
}

ModuleData::ModuleData( const ModuleData *d )
{
   off_file_name = d->off_file_name;
   mesh_unit = d->mesh_unit;

   cp_x = d->cp_x;
   cp_y = d->cp_y;
   cp_z = d->cp_z;
   theta = d->theta;
   phi = d->phi;
   psi = d->psi;

   if ( d->grf )
      grf = new GREEN( *d->grf );
   else
      grf = NULL;

   plist = d->plist;
   tlist = d->tlist;
   velist = d->velist;


   J_probe_file_name = d->J_probe_file_name;

   startpkt[0] = d->startpkt[0];
   startpkt[1] = d->startpkt[1];
   startpkt[2] = d->startpkt[2];
   noofpkt = d->noofpkt;
   endpkt[0] = d->endpkt[0];
   endpkt[1] = d->endpkt[1];
   endpkt[2] = d->endpkt[2];

   extract_psi = d->extract_psi;

   setup_exvec();
}

ModuleData::~ModuleData()
{
   delete grf;
}



class SORTpoint
{
   public:
      int no;
      double x, y, z;

      SORTpoint( int no_in=-1, double x_in=0.0, double y_in=0.0, double z_in=0.0 )
      : no(no_in), x(x_in), y(y_in), z(z_in)
      {}

      bool operator> ( const SORTpoint& sp2 ) const
      {
         if ( x > sp2.x )
            return(true);
         else if ( x == sp2.x )
            if ( y > sp2.y )
               return(true);
            else if ( y == sp2.y )
               if ( z > sp2.z )
                  return(true);

/**** doesn't work
         if ( x - sp2.x > DBL_EPSILON )
            return(true);
         else if ( y - sp2.y > DBL_EPSILON )   
            return(true);
         else if ( z > sp2.z )   
            return(true);
****/

         return(false);
      };

      bool operator< ( const SORTpoint& sp2 ) const
      {
         if ( x < sp2.x )
            return(true);
         else if ( x == sp2.x )
            if ( y < sp2.y )   
               return(true);   
            else if ( y == sp2.y )
               if ( z < sp2.z )
                  return(true);

/**** doesn't work
         if ( x - sp2.x < -DBL_EPSILON )
            return(true);
         else if ( y - sp2.y < -DBL_EPSILON )
            return(true);
         else if ( z < sp2.z )
            return(true);
****/

         return(false);
      };
};

class SORTdata
{
   public:
      int p1, p2;
      int p3, t;

      SORTdata( int p1_in=-1, int p2_in=-1, int p3_in=-1, int t_in=-1 )
      : p1(p1_in), p2(p2_in), p3(p3_in), t(t_in)
      {}

      bool operator> ( const SORTdata& sd2 ) const
      {
         if ( p1 > sd2.p1 )
            return(true);
         else if ( p1 == sd2.p1 )
            if ( p2 > sd2.p2 )
               return(true);
         return(false);
      };

      bool operator< ( const SORTdata& sd2 ) const
      {
         if ( p1 < sd2.p1 )
            return(true);
         else if ( p1 == sd2.p1 )
            if ( p2 < sd2.p2 )
               return(true);
         return(false);
      };
};

void ModuleData::read_off_file()
{
   char cbuff;
   int pno_t, tno_t, eno_t, n;
   double x, y, z, rmat[3][3];
   string strbuff;
   ifstream off_file;
   SORTpoint sp_t;
   TRIANGLE t_t;
   vector<SORTpoint> splist;

   splist.clear();

   plist.clear();
   tlist.clear();
   velist.clear();

   rotmat( rmat, mesh_unit, psi, 0.0, 0.0, theta, phi );

   off_file.open( off_file_name.c_str(), ios::in );
   if ( !off_file.is_open() )
   {
      throw( "cannot open OFF file" );
   }

// head
   off_file >> strbuff;

   if ( strbuff != "OFF" )
   {
      throw( "not an OFF file" );
   }
   off_file >> strbuff;
   pno_t = atoi(strbuff.c_str());
   off_file >> strbuff;
   tno_t = atoi(strbuff.c_str());
   off_file >> strbuff;
   eno_t = atoi(strbuff.c_str());

// points
   for ( int p=0; p<pno_t; p++ )
   {
      off_file >> strbuff;
      x = atof(strbuff.c_str());
      off_file >> strbuff;
      y = atof(strbuff.c_str());
      off_file >> strbuff;
      z = atof(strbuff.c_str());

      sp_t.no = p;
      sp_t.x = rmat[0][0]*x + rmat[0][1]*y + rmat[0][2]*z + cp_x;
      sp_t.y = rmat[1][0]*x + rmat[1][1]*y + rmat[1][2]*z + cp_y;
      sp_t.z = rmat[2][0]*x + rmat[2][1]*y + rmat[2][2]*z + cp_z;

      splist.push_back( sp_t );
   }

// triangles
   for ( int t=0; t<tno_t; t++ )
   {
      off_file >> strbuff;
      n = atoi(strbuff.c_str());
      if ( n == 3 )
      {
         off_file >> strbuff;
         t_t.p1 = atoi(strbuff.c_str());
         off_file >> strbuff;
         t_t.p2 = atoi(strbuff.c_str());
         off_file >> strbuff;
         t_t.p3 = atoi(strbuff.c_str());

         t_t.ve1 = -1;
         t_t.ve2 = -1;
         t_t.ve3 = -1;
         t_t.sign.sp1 = 0;
         t_t.sign.sp2 = 0;
         t_t.sign.sp3 = 0;

         if ( t_t.p1 < 0 ||
              t_t.p2 < 0 ||
              t_t.p3 < 0 ||
              (unsigned)abs(t_t.p1) >= splist.size() ||
              (unsigned)abs(t_t.p2) >= splist.size() ||
              (unsigned)abs(t_t.p3) >= splist.size()
            )
         {
            throw( "wrong point index" );
         }

         tlist.push_back( t_t );
      }
      else
      {
         throw( "non-triangle found" );
      }

      off_file.get(cbuff);
      while ( cbuff != '\n' && !off_file.eof() )
      {
         off_file.get(cbuff); 
      }

   }

   if ( off_file.fail() )
   {
      throw( "trouble with OFF file" );
   }

   off_file.close();

   if ( tlist.size() < 1 )
   {
      throw( "no triangles found" );
   }


//
// remove multiple points
//
   sort( splist.begin(), splist.end() );

   plist.clear();

   POINT p_t;
   int *permute, p;
   permute = new int[ splist.size() ];

   vector<SORTpoint>::iterator isp1, isp2;
   isp1 = splist.begin();

   p = 0;
   p_t.x = isp1->x;
   p_t.y = isp1->y;
   p_t.z = isp1->z;
   plist.push_back( p_t );
   permute[isp1->no] = p;

   for ( isp2 = isp1 + 1;
         isp2 != splist.end();
         isp1++, isp2++ )
   {
      if ( HYPOT( isp1->x-isp2->x, isp1->y-isp2->y, isp1->z-isp2->z ) > DIST_EPSILON )
      {
         p++;
         p_t.x = isp2->x;
         p_t.y = isp2->y;
         p_t.z = isp2->z;
         plist.push_back( p_t );
         permute[isp2->no] = p;
      }
      else
      {
         permute[isp2->no] = p;
      }
   }

   for ( TRIANGLE_LIST::iterator it = tlist.begin();
         it != tlist.end();
         it++ )
   {
      it->p1 = permute[it->p1];
      it->p2 = permute[it->p2];
      it->p3 = permute[it->p3];
   }

   delete [] permute;
   splist.clear();


//
// searching edges
//
   int t, veno;
   vector<SORTdata> SDlist;
   vEDGE vE_t;
   vector<SORTdata>::iterator sd1, sd2;

   veno = 0;

   SDlist.clear();

   for ( TRIANGLE_LIST::iterator it = tlist.begin();
         it != tlist.end();
         it++ )
   {
      t = it - tlist.begin();

// edge 1
      if ( it->p1 < it->p2 )
      {
         SDlist.push_back( SORTdata( it->p1, it->p2, it->p3, t ) );
      }
      else
      {
         SDlist.push_back( SORTdata( it->p2, it->p1, it->p3, t ) );
      }

// edge 2
      if ( it->p2 < it->p3 )
      {
         SDlist.push_back( SORTdata( it->p2, it->p3, it->p1, t ) );
      }
      else
      {
         SDlist.push_back( SORTdata( it->p3, it->p2, it->p1, t ) );
      }

// edge 3
      if ( it->p3 < it->p1 )
      {
         SDlist.push_back( SORTdata( it->p3, it->p1, it->p2, t ) );
      }
      else
      {
         SDlist.push_back( SORTdata( it->p1, it->p3, it->p2, t ) );
      }

   }


   sort( SDlist.begin(), SDlist.end() );


   sd1=SDlist.begin();
   while( sd1->p1<0 || sd1->p2<0 || sd1->p3<0 || sd1->t<0 ) sd1++;
   for ( sd2 = sd1 + 1;
         sd2 != SDlist.end();
         sd1++, sd2++ )
   {
      if ( sd1->p1 == sd2->p1 && sd1->p2 == sd2->p2 )
      {
         vE_t.p1 = sd1->p1;
         vE_t.p2 = sd1->p2;
         vE_t.spm = sd1->p3;
         vE_t.spp = sd2->p3;
         vE_t.trim = sd1->t;
         vE_t.trip = sd2->t;
         velist.push_back( vE_t );

         if ( tlist[sd1->t].p1 == sd1->p3 )
         {
            tlist[sd1->t].ve1  = veno;
            tlist[sd1->t].sign.sp1 = -1;
         }
         else if ( tlist[sd1->t].p2 == sd1->p3 )
         {
            tlist[sd1->t].ve2  = veno;
            tlist[sd1->t].sign.sp2 = -1;
         }
         else if ( tlist[sd1->t].p3 == sd1->p3 )
         {
            tlist[sd1->t].ve3  = veno;
            tlist[sd1->t].sign.sp3 = -1;
         }
         else
         {
            cerr << "Error in ModuleData::ModuleData(): internal line "
                 << __LINE__ << "!" << endl;
            exit(1);
         }

         if ( tlist[sd2->t].p1 == sd2->p3 )
         {
            tlist[sd2->t].ve1  = veno;
            tlist[sd2->t].sign.sp1 = +1;
         }
         else if ( tlist[sd2->t].p2 == sd2->p3 )
         {
            tlist[sd2->t].ve2  = veno;
            tlist[sd2->t].sign.sp2 = +1;
         }
         else if ( tlist[sd2->t].p3 == sd2->p3 )
         {
            tlist[sd2->t].ve3  = veno;
            tlist[sd2->t].sign.sp3 = +1;
         }
         else
         {
            cerr << "Error in ModuleData::ModuleData(): internal line "
                 << __LINE__ << "!" << endl;
            exit(1);
         }

         sd1++, sd2++, veno++;
      }
   }

   SDlist.clear();
}


#ifdef HAVE_LIBGL
void ModuleData::ogl_triangles( GLuint list )
{
   const GLfloat clr_1[4] = { 0.12, 0.90, 0.90, 1.0 };
   const GLfloat clr_2[4] = { 0.78, 0.20, 0.12, 1.0 };

   double u[3], v[3], w[3], a;

   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, clr_1 );
   glMaterialfv( GL_BACK, GL_AMBIENT_AND_DIFFUSE, clr_2 );

   glBegin(GL_TRIANGLES);

   for ( vector<TRIANGLE>::iterator it = tlist.begin();
         it != tlist.end();  
         it++ )
   {

      u[0] = plist[ it->p2 ].x -  plist[ it->p1 ].x;
      u[1] = plist[ it->p2 ].y -  plist[ it->p1 ].y;
      u[2] = plist[ it->p2 ].z -  plist[ it->p1 ].z;

      v[0] = plist[ it->p3 ].x -  plist[ it->p2 ].x;
      v[1] = plist[ it->p3 ].y -  plist[ it->p2 ].y;
      v[2] = plist[ it->p3 ].z -  plist[ it->p2 ].z;

      w[0] = u[1]*v[2] - u[2]*v[1];
      w[1] = u[2]*v[0] - u[0]*v[2];
      w[2] = u[0]*v[1] - u[1]*v[0];

      a = sqrt( w[0]*w[0] + w[1]*w[1] + w[2]*w[2] );

      glNormal3f( w[0]/a, w[1]/a, w[2]/a );


      glVertex3f( plist[it->p1].x,  plist[it->p1].y, plist[it->p1].z );
      glVertex3f( plist[it->p2].x,  plist[it->p2].y, plist[it->p2].z );
      glVertex3f( plist[it->p3].x,  plist[it->p3].y, plist[it->p3].z );
   }

   glEnd();
}
#endif


void ModuleData::triangle_center( int tno, POINT &cp ) const
{
#ifdef BOUNDARY_CHECK
   if ( tno < 0 || tno >= (int)tlist.size() )
   {
      cerr << "Error in ModuleData::triangle_center(): tno out of bounds!"
           << endl;
      exit(1);
   }
#endif
   cp.x = ( plist[ tlist[tno].p1 ].x
            + plist[ tlist[tno].p2 ].x
              + plist[ tlist[tno].p3 ].x ) / 3.0;
   cp.y = ( plist[ tlist[tno].p1 ].y
            + plist[ tlist[tno].p2 ].y
              + plist[ tlist[tno].p3 ].y ) / 3.0;
   cp.z = ( plist[ tlist[tno].p1 ].z
            + plist[ tlist[tno].p2 ].z
              + plist[ tlist[tno].p3 ].z ) / 3.0;
}

bool ModuleData::triangle_neigh( int tno1, int tno2 ) const
{
   POINT cp1, cp2;
   double r1, r2, R;

   triangle_center( tno1, cp1 );
   triangle_center( tno2, cp2 );

   r1 = HYPOT( cp1.x-plist[ tlist[tno1].p1 ].x,
               cp1.y-plist[ tlist[tno1].p1 ].y,
               cp1.z-plist[ tlist[tno1].p1 ].z  );
   R = HYPOT( cp1.x-plist[ tlist[tno1].p2 ].x,
              cp1.y-plist[ tlist[tno1].p2 ].y,
              cp1.z-plist[ tlist[tno1].p2 ].z  );
   r1 = ( R > r1 ? R : r1 );
   R = HYPOT( cp1.x-plist[ tlist[tno1].p3 ].x,
              cp1.y-plist[ tlist[tno1].p3 ].y,
              cp1.z-plist[ tlist[tno1].p3 ].z  );
   r1 = ( R > r1 ? R : r1 );

   r2 = HYPOT( cp2.x-plist[ tlist[tno2].p1 ].x,
               cp2.y-plist[ tlist[tno2].p1 ].y,
               cp2.z-plist[ tlist[tno2].p1 ].z  );
   R = HYPOT( cp2.x-plist[ tlist[tno2].p2 ].x,
              cp2.y-plist[ tlist[tno2].p2 ].y,
              cp2.z-plist[ tlist[tno2].p2 ].z  );
   r2 = ( R > r2 ? R : r2 );
   R = HYPOT( cp2.x-plist[ tlist[tno2].p3 ].x,
              cp2.y-plist[ tlist[tno2].p3 ].y,
              cp2.z-plist[ tlist[tno2].p3 ].z  ); 
   r2 = ( R > r2 ? R : r2 );

   if ( HYPOT( cp1.x-cp2.x, cp1.y-cp2.y, cp1.z-cp2.z ) > r1+r2 ) return(false);

   return(true);
}


double ModuleData::triangle_area( int tno ) const
{
   double nx, ny, nz;
   return( triangle_normal( tno, nx, ny, nz ) );
}


int ModuleData::near_triangle( POINT &p ) const
{
   int nt = 0;
   POINT tp;

   triangle_center( nt, tp );

   double dt, d = HYPOT( p.x-tp.x, p.y-tp.y, p.z-tp.z );

   for ( int t=1; t<(int)tlist.size(); t++ )
   {
      triangle_center( t, tp );
      if ( d > ( dt = HYPOT( p.x-tp.x, p.y-tp.y, p.z-tp.z ) ) )
      {
         d = dt;
         nt = t;
      }
   }

   return(nt);
}


void ModuleData::setup_exvec()
{
   double rmat[3][3];
   double theta, phi, x, y, z, rho;

   x = endpkt[0] - startpkt[0];
   y = endpkt[1] - startpkt[1];
   z = endpkt[2] - startpkt[2];
   rho = hypot( x, y );

   theta = atan2( rho, z );
   phi = atan2( y, x );

   rotmat( rmat, 1.0, extract_psi, 0.0, 0.0, theta, phi );

   lv[0] = rmat[0][2];
   lv[1] = rmat[1][2];
   lv[2] = rmat[2][2];

   c1v[0] = rmat[0][0];
   c1v[1] = rmat[1][0];
   c1v[2] = rmat[2][0];

   c2v[0] = rmat[0][1];
   c2v[1] = rmat[1][1];
   c2v[2] = rmat[2][1];
}


double ModuleData::triangle_normal( int tno, double &nx, double &ny, double &nz ) const
{
   double a[3], b[3], AR2;

#ifdef BOUNDARY_CHECK
   if ( tno < 0 || tno >= (int)tlist.size() )
   {
      cerr << "Error in ModuleData::triangle_normal(): tno out of bounds!"
           << endl;
      exit(1);
   }
#endif

   a[0] = plist[ tlist[tno].p2 ].x - plist[ tlist[tno].p1 ].x;
   a[1] = plist[ tlist[tno].p2 ].y - plist[ tlist[tno].p1 ].y;
   a[2] = plist[ tlist[tno].p2 ].z - plist[ tlist[tno].p1 ].z;

   b[0] = plist[ tlist[tno].p3 ].x - plist[ tlist[tno].p1 ].x;
   b[1] = plist[ tlist[tno].p3 ].y - plist[ tlist[tno].p1 ].y;
   b[2] = plist[ tlist[tno].p3 ].z - plist[ tlist[tno].p1 ].z;

   nx = a[1]*b[2] - a[2]*b[1];
   ny = a[2]*b[0] - a[0]*b[2];
   nz = a[0]*b[1] - a[1]*b[0];

   AR2 = HYPOT( nx, ny, nz );
   nx /= AR2;
   ny /= AR2;
   nz /= AR2;

   return( AR2 * 0.5 );
}

