//
//  Routines for VECT input/output
//  (c) 19.11.2003 Ingo D. Rullhusen
//

#include <iostream>
#include <fstream>
#include <vector>
#include <stdlib.h>
#include <math.h>

#include "vect.h"


POLY::POLY()
{
   nv = 0;
   color[0] = color[1] = color[2] = -1;
   x.clear();
   y.clear();
   z.clear();
}


POLY::POLY( int nvi, int r, int g, int b )
{
   nv = nvi;
   color[0] = r;
   color[1] = g;
   color[2] = b;
   if ( nv != 0 )
   {
      x.resize( abs(nv) );
      y.resize( abs(nv) );
      z.resize( abs(nv) );
      for ( int i=0; i<abs(nv); i++ )
      {
         x[i] = 0.0;
         y[i] = 0.0;
         z[i] = 0.0;
      }
   }
   else
   {
      x.clear();
      y.clear();
      z.clear();
   }
}


POLY::~POLY(){}


void vect_out( const string filename, POLYLIST *plist )
{
   int npoly, nvertices, ncolors, i;
   POLYLIST::iterator ipo;
   ofstream vect;

   npoly = 0;
   nvertices = 0;
   ncolors = 0;
   for ( ipo=plist->begin(); ipo!=plist->end(); ipo++ )
   {
      if ( abs(ipo->nv) > 0 )
      {
         npoly++;
         nvertices += abs(ipo->nv);
         if ( ipo->color[0] > -1 && ipo->color[1] > -1 && ipo->color[2] > -1 )
            ncolors += 1;
      }
   }

   vect.open( filename.c_str(), ios::out );
   vect << "VECT" << endl;
   vect << npoly << " " << nvertices << " " << ncolors << endl;

// Nv
   for ( ipo=plist->begin(); ipo!=plist->end(); ipo++ )
   {
      if ( abs(ipo->nv) > 0 )
         vect << ipo->nv << " ";
   }
   vect << endl;

// Nc
   for ( ipo=plist->begin(); ipo!=plist->end(); ipo++ )
   {
      if ( abs(ipo->nv) > 0 )
         if ( ipo->color[0] < 0 || ipo->color[1] < 0 || ipo->color[2] < 0 )
            vect << "0 ";
         else
            vect << "1 ";
   }
   vect << endl;

// Vert
   for ( ipo=plist->begin(); ipo!=plist->end(); ipo++ )
   {
      for ( i=0; i<abs(ipo->nv); i++ )
         vect << ipo->x[i] << " " << ipo->y[i] << " " << ipo->z[i] << " ";
   }
   vect << endl;

// Color
   for ( ipo=plist->begin(); ipo!=plist->end(); ipo++ )
   {
      if ( abs(ipo->nv) > 0 &&
           ipo->color[0] > -1 && ipo->color[1] > -1 && ipo->color[2] > -1 )
      {
         vect << ipo->color[0]/255.0 << " "
              << ipo->color[1]/255.0 << " "
              << ipo->color[2]/255.0 << " "
              << "0.0" << " ";
      }
   }
   vect << endl;

   vect.close();
}


void delete_poly_list( POLYLIST *plist )
{
   plist->clear();
}


int vect_in( const string filename, POLYLIST *plist )
{
   char header[4];
   int npoly, nvertices, ncolors, *nc, i, j;
   double r, g, b, a;
   POLYLIST::iterator ipo;
   ifstream vect;

   vect.open( filename.c_str(), ios::in );

   if ( !vect.is_open() ) return(1);

   plist->clear();

// Header
   vect >> header[0] >> header[1] >> header[2] >> header[3];
   if ( header[0] != 'V' || header[1] != 'E' ||
        header[2] != 'C' || header[3] != 'T'    ) return(2);

// Size
   vect >> npoly >> nvertices >> ncolors;
   if ( !npoly ) return(0);

// Nv
   plist->resize(npoly);
   for ( ipo=plist->begin(); ipo!=plist->end(); ipo++ )
   {
      if ( vect.eof() ) return(3);
      vect >> ipo->nv;
      ipo->x.resize(abs(ipo->nv));
      ipo->y.resize(abs(ipo->nv));
      ipo->z.resize(abs(ipo->nv));
   }

// Nc
   nc = new int[npoly];
   for ( i=0; i<npoly; i++ ) vect >> nc[i];

// Vert
   for ( ipo=plist->begin(); ipo!=plist->end(); ipo++ )
   {
      for ( i=0; i<abs(ipo->nv); i++ )
      {
         if ( vect.eof() )
         {
            delete [] nc;
            return(3);
         }
         vect >> ipo->x[i] >> ipo->y[i] >> ipo->z[i];
      }
   }

// Color
   for ( ipo=plist->begin(), j=0; ipo!=plist->end(); ipo++, j++ )
   {
      r = g = b = -1.0 / 255.0;
#ifdef BOUNDARY_CHECK
      if ( j >= npoly )
      {
         cerr << "Error in vect_in(): internal." << endl;
         exit(1);
      }
#endif
      for ( i=0; i<nc[j]; i++ )
      {
         if ( vect.eof() )
         {
            delete [] nc;
            return(3);
         }
         vect >> r >> g >> b >> a;
      }
      ipo->color[0] = (int)rint(r*255.0);
      ipo->color[1] = (int)rint(g*255.0);
      ipo->color[2] = (int)rint(b*255.0);
   }

   delete [] nc;

   return(0);
}


POLY* append_ch( const int color[3],
                 const double bp[3], const double ln[3], const double hgt[3],
                 const int ch_no, const double *ch_x, const double *ch_y,
                 POLYLIST *plist )
{
   int i;
   POLY nhead;

   nhead.color[0] = color[0];
   nhead.color[1] = color[1];
   nhead.color[2] = color[2];

   nhead.nv = ch_no;

   nhead.x.resize(abs(ch_no));
   nhead.y.resize(abs(ch_no));
   nhead.z.resize(abs(ch_no));

   for ( i=0; i<abs(ch_no) ; i++ )
   {
      nhead.x[i] = bp[0] + ln[0] * ch_x[i] + hgt[0] * ch_y[i];
      nhead.y[i] = bp[1] + ln[1] * ch_x[i] + hgt[1] * ch_y[i];
      nhead.z[i] = bp[2] + ln[2] * ch_x[i] + hgt[2] * ch_y[i];
   }

   plist->push_back(nhead);

   return(&(*(plist->end()-1)));
}


void vect_num( const double cp[3], const double ln[3], const double hgt[3],
               const int color[3], int no, POLYLIST *plist )
{
   int nextint, fig, base, i;
   double bp[3];

// size
   const double width = 2.0;
//   const double height = 3.0;
// sign
   const int sign_no = 2;
   const double sign_x[2] = { 0.8, 1.8 };
   const double sign_y[2] = { 1.5, 1.5 };
// one
   const int one_no = 3;
   const double one_x[3] = { 0.5, 1.3, 1.3 };
   const double one_y[3] = { 1.6, 3.0, 0.0 };
// two
   const int two_no = 8;
   const double two_x[8] = { 0.3, 0.3, 0.6, 1.4, 1.7, 1.7, 0.3, 1.7 };
   const double two_y[8] = { 2.4, 2.7, 3.0, 3.0, 2.7, 2.4, 0.0, 0.0 };
// three
   const int three_no = 10;
   const double three_x[10] = { 0.3, 0.3, 1.7, 0.9, 1.4, 1.7, 1.7, 1.4, 0.6, 0.3 };
   const double three_y[10] = { 2.7, 3.0, 3.0, 1.5, 1.5, 1.2, 0.3, 0.0, 0.0, 0.3 };
// four
   const int four_no = 4;
   const double four_x[4] = { 1.3, 1.3, 0.3, 1.7 };
   const double four_y[4] = { 0.0, 3.0, 1.1, 1.1 };
// five
   const int five_no = 11;
   const double five_x[11] = { 1.7, 1.7, 0.3, 0.3, 0.6, 1.4, 1.7, 1.7, 1.4, 0.6, 0.3 };
   const double five_y[11] = { 2.7, 3.0, 3.0, 1.2, 1.5, 1.5, 1.2, 0.3, 0.0, 0.0, 0.3 };
// six
   const int six_no = 12;
   const double six_x[12] = { 1.7, 1.4, 0.7, 0.3, 0.3, 0.6, 1.4, 1.7, 1.7, 1.4, 0.6, 0.3 };
   const double six_y[12] = { 2.7, 3.0, 3.0, 2.1, 0.3, 0.0, 0.0, 0.3, 1.2, 1.5, 1.5, 1.2 };
// seven
   const int seven_no = 5;
   const double seven_x[5] = { 0.3, 0.3, 1.7, 0.6, 0.6 };
   const double seven_y[5] = { 2.7, 3.0, 3.0, 1.0, 0.0 };
// eight
   const int eight_no = -16;
   const double eight_x[16] = { 1.4, 0.6, 0.3, 0.3, 0.6, 1.4, 1.7, 1.7, 1.4, 0.6, 0.3, 0.3, 0.6, 1.4, 1.7, 1.7 };
   const double eight_y[16] = { 3.0, 3.0, 2.7, 1.8, 1.5, 1.5,1.2, 0.3, 0.0, 0.0, 0.3, 1.2, 1.5, 1.5, 1.8, 2.7 };
// nine
   const int nine_no = 12;
   const double nine_x[12] = { 1.7, 1.4, 0.6, 0.3, 0.3, 0.6, 1.4, 1.7, 1.7, 1.3, 0.6, 0.3 };
   const double nine_y[12] = { 1.8, 1.5, 1.5, 1.8, 2.7, 3.0, 3.0, 2.7, 0.9, 0.0, 0.0, 0.3 };
// zero
   const int zero_no = -8;
   const double zero_x[8] = { 0.6, 0.3, 0.3, 0.6, 1.4, 1.7, 1.7, 1.4 };
   const double zero_y[8] = { 3.0, 2.7, 0.3, 0.0, 0.0, 0.3, 2.7, 3.0 };
// cbar
   const int cbar_no = 2;
   const double cbar_x[2] = { 0.6, 1.2 };
   const double cbar_y[2] = { 1.5, 1.5 };

   bp[0] = cp[0];
   bp[1] = cp[1];
   bp[2] = cp[2];

   if ( no < 0 )
   {
      no *= -1;
      append_ch( color, bp, ln, hgt, sign_no, sign_x, sign_y, plist );

      bp[0] += width * ln[0];
      bp[1] += width * ln[1];
      bp[2] += width * ln[2];
   }

   nextint = no;
   i = ( nextint == 0 ? 0 : (int)floor(log10((double)nextint)) );

   while ( i > -1 )
   {
      base = (int)( pow( 10.0, i ) );
      fig = nextint / base;
      switch( fig )
      {
         case 0:
            append_ch( color, bp, ln, hgt, zero_no, zero_x, zero_y, plist );
            break;
         case 1:
            append_ch( color, bp, ln, hgt, one_no, one_x, one_y, plist );
            break;
         case 2:
            append_ch( color, bp, ln, hgt, two_no, two_x, two_y, plist );   
            break;
         case 3:
            append_ch( color, bp, ln, hgt, three_no, three_x, three_y, plist );
            break;
         case 4:
            append_ch( color, bp, ln, hgt, four_no, four_x, four_y, plist );
            break;
         case 5:
            append_ch( color, bp, ln, hgt, five_no, five_x, five_y, plist );
            break;
         case 6:
            append_ch( color, bp, ln, hgt, six_no, six_x, six_y, plist );
            break;
         case 7:
            append_ch( color, bp, ln, hgt, seven_no, seven_x, seven_y, plist );
            append_ch( color, bp, ln, hgt, cbar_no, cbar_x, cbar_y, plist );
            break;
         case 8:
            append_ch( color, bp, ln, hgt, eight_no, eight_x, eight_y, plist );
            break;
         case 9:
            append_ch( color, bp, ln, hgt, nine_no, nine_x, nine_y, plist );
            break;
      }

      bp[0] += width * ln[0];
      bp[1] += width * ln[1];
      bp[2] += width * ln[2];

      nextint -= fig * base;
      i--;
   }

}


void vect_harrow( const double p[2], const double d,
                  const double rho, const double phi,
                  int r, int g, int b, POLYLIST *plist )
{
   const double arrowangle = 150.0 * M_PI/180.0;
   const double arrowhead  = 0.25;
   POLY *po;

   po = new POLY( 2, r, g, b );

   po->x[0] = p[0];
   po->y[0] = p[1];
   po->z[0] = d;

   po->x[1] = po->x[0] + rho*cos(phi);
   po->y[1] = po->y[0] + rho*sin(phi);
   po->z[1] = d;

   plist->push_back(*po);

   delete po;

   po = new POLY( 3, r, g, b );

   po->x[1] = p[0] + rho*cos(phi);
   po->y[1] = p[1] + rho*sin(phi);
   po->z[1] = d;

   po->x[0] = po->x[1] + rho*arrowhead*cos( phi - arrowangle );
   po->y[0] = po->y[1] + rho*arrowhead*sin( phi - arrowangle );
   po->z[0] = d;

   po->x[2] = po->x[1] + rho*arrowhead*cos( phi + arrowangle );
   po->y[2] = po->y[1] + rho*arrowhead*sin( phi + arrowangle );
   po->z[2] = d;

   plist->push_back(*po);

   delete po;
}
