//
// Class for field grid
// (c) 13.3.2007 Ingo D. Rullhusen
//

#include <iostream>

#include "aimfield.h"


static const complex<double> cNULL(0.0,0.0);


AIMfield::AIMfield( const int nx, const int ny, const int nz )
: nox(nx), noy(ny), noz(nz)
{
   noxy = nox*noy;
   noxyz = noxy*noz;

   transformed = false;

   X = new complex<double>[noxyz];
   Y = new complex<double>[noxyz];
   Z = new complex<double>[noxyz];
   div = new complex<double>[noxyz];

#pragma omp critical (FFTW)
{
   pX = fftw_plan_dft_3d( noz, noy, nox,
                          (fftw_complex*)X, (fftw_complex*)X,
                          FFTW_BACKWARD, FFTW_ESTIMATE
                        );

   pY = fftw_plan_dft_3d( noz, noy, nox,
                          (fftw_complex*)Y, (fftw_complex*)Y,
                          FFTW_BACKWARD, FFTW_ESTIMATE
                        );

   pZ = fftw_plan_dft_3d( noz, noy, nox,
                          (fftw_complex*)Z, (fftw_complex*)Z,
                          FFTW_BACKWARD, FFTW_ESTIMATE
                        );

   pdiv = fftw_plan_dft_3d( noz, noy, nox,
                            (fftw_complex*)div, (fftw_complex*)div,
                            FFTW_BACKWARD, FFTW_ESTIMATE
                          );


   IpX = fftw_plan_dft_3d( noz, noy, nox,
                           (fftw_complex*)X, (fftw_complex*)X,
                           FFTW_FORWARD, FFTW_ESTIMATE
                         );

   IpY = fftw_plan_dft_3d( noz, noy, nox,
                           (fftw_complex*)Y, (fftw_complex*)Y,
                            FFTW_FORWARD, FFTW_ESTIMATE
                         );

   IpZ = fftw_plan_dft_3d( noz, noy, nox,
                           (fftw_complex*)Z, (fftw_complex*)Z,
                           FFTW_FORWARD, FFTW_ESTIMATE
                         );

   Ipdiv = fftw_plan_dft_3d( noz, noy, nox,
                             (fftw_complex*)div, (fftw_complex*)div,
                             FFTW_FORWARD, FFTW_ESTIMATE
                           );
}  // end:pragma omp critical (FFTW)

}

AIMfield::~AIMfield()
{
   delete [] X;
   delete [] Y;
   delete [] Z;
   delete [] div;

#pragma omp critical (FFTW)
{
   fftw_destroy_plan(pdiv);
   fftw_destroy_plan(pZ);
   fftw_destroy_plan(pY);
   fftw_destroy_plan(pX);

   fftw_destroy_plan(Ipdiv);
   fftw_destroy_plan(IpZ);
   fftw_destroy_plan(IpY);
   fftw_destroy_plan(IpX);
}  // end:omp critical (FFTW)
}


void AIMfield::FFT()
{
#ifdef BOUNDARY_CHECK
   if ( transformed )
   {
      cerr << "Error in AIMfield::FFT(): FFT of transformed field!"
           << endl;
      exit(1);
   }
#endif
   transformed = true;

   fftw_execute(pX);
   fftw_execute(pY);
   fftw_execute(pZ);
   fftw_execute(pdiv);
}

void AIMfield::IFFT()
{
#ifdef BOUNDARY_CHECK
   if ( !transformed )
   {
      cerr << "Error in AIMfield::IFFT(): IFFT of untransformed field!"
           << endl;
      exit(1);
   }
#endif
   transformed = false;

   fftw_execute(IpX);
   fftw_execute(IpY);
   fftw_execute(IpZ);
   fftw_execute(Ipdiv);

   double inv = 1.0 / noxyz;
   for ( int i=0; i<noxyz; i++ )
   {
      X[i] *= inv;
      Y[i] *= inv;
      Z[i] *= inv;
      div[i] *= inv;
   }
}


void AIMfield::mult_lambda( complex<double> *vp, complex<double> *phi,
                            AIMData *ad )
{
   int id_a, id_ay, id_az, id_l;

#ifdef BOUNDARY_CHECK
   if ( transformed )
   {
      cerr << "Error in AIMfield::mult_lambda(): transformed field!"
           << endl;
      exit(1);
   }
#endif

   *vp = *phi = cNULL;

   for ( int k=0; k<ad->noz; k++ )
   {
      id_az = nox*noy * ( k + ad->z_off );
      for ( int j=0; j<ad->noy; j++ )
      {
         id_ay = id_az + nox * ( j + ad->y_off );
         for ( int i=0; i<ad->nox; i++ )
         {
            id_a = id_ay + ( i + ad->x_off );
#ifdef BOUNDARY_CHECK
            if ( id_a >= noxyz )
            {
               cerr << "Error in AIMfield::mult_lambda(): index out of bounds!"
                    << endl;
               exit(1);
            }
#endif
            id_l = ad->index( i, j, k );
            *vp += (double)ad->lambda_x[id_l] * X[ id_a ]
                   + (double)ad->lambda_y[id_l] * Y[ id_a ]
                   + (double)ad->lambda_z[id_l] * Z[ id_a ];

            *phi += (double)ad->lambda_div[id_l] * div[ id_a ];
         }
      }
   }

}


void AIMfield::set2zero( bool transformed_ )
{
   transformed = transformed_;

   for ( int i=0; i<noxyz; i++ )
   {
      X[i] = cNULL;
      Y[i] = cNULL;
      Z[i] = cNULL;
      div[i] = cNULL;
   }
}

void AIMfield::add( complex<double> fac, AIMData ad )
{
   int id_a, id_ay, id_az, id_l;

#ifdef BOUNDARY_CHECK
   if ( transformed )
   {
      cerr << "Error in AIMfield::add(): transformed field!"
           << endl;
      exit(1);
   }
#endif

   for ( int k=0; k<ad.noz; k++ )
   {
      id_az = nox*noy * ( k + ad.z_off );
      for ( int j=0; j<ad.noy; j++ )
      {
         id_ay = id_az + nox * ( j + ad.y_off );
         for ( int i=0; i<ad.nox; i++ )
         {
            id_a = id_ay + ( i + ad.x_off );
#ifdef BOUNDARY_CHECK
            if ( id_a >= noxyz )
            {
               cerr << "Error in AIMfield::add(): index out of bounds!"
                    << endl;
               exit(1);
            }
#endif
            id_l = ad.index( i, j, k );
            X[ id_a ] += fac * (double)ad.lambda_x[id_l];
            Y[ id_a ] += fac * (double)ad.lambda_y[id_l];
            Z[ id_a ] += fac * (double)ad.lambda_z[id_l];
            div[ id_a ] += fac * (double)ad.lambda_div[id_l];
         }
      }
   }


}
