//
// Class for AIM solver
// (c) 21.8.2006 Ingo D. Rullhusen
//

#include <iostream>

#include <qprogressdialog.h>

#include "natconst.h"
#include "aimsolver.h"
#include "auxiliary_qt.h"
#include "fieldx.h"

#include "global_vars.h"


static const complex<double> cNULL(0.0,0.0);
static const complex<double> cJ(0.0,1.0);


AIMsolver::AIMsolver()
{
   veno = 0;
   RHS = J = NULL;

   itno = 0;
   epsilon = 0.0;

   no_of_inc_modules = 0;
   inc_modno = NULL;
   start_idx_inc_mod = NULL;
   ADinc.clear();
   AMPLinc.clear();

   ff = NULL;
}

AIMsolver::~AIMsolver()
{
   clear();
}


void AIMsolver::read_binary( fstream &file )
{
   delete [] RHS;
   delete [] J;

   delete [] inc_modno;
   delete [] start_idx_inc_mod;
   ADinc.clear();
   AMPLinc.clear();


   file.read( (char *) &veno, sizeof(veno) );

   RHS = new complex<double>[veno];
   J   = new complex<double>[veno];

   file.read( (char *) RHS, veno*sizeof(*RHS) );
   file.read( (char *)   J, veno*sizeof(*J)   );


   file.read( (char *) &itno, sizeof(itno) );
   file.read( (char *) &epsilon, sizeof(epsilon) );


   file.read( (char *) &no_of_inc_modules, sizeof(no_of_inc_modules) );

   inc_modno = new int[no_of_inc_modules];
   start_idx_inc_mod = new int[ no_of_inc_modules + 1 ];

   file.read( (char *)         inc_modno,  no_of_inc_modules   *sizeof(*inc_modno) );
   file.read( (char *) start_idx_inc_mod, (no_of_inc_modules+1)*sizeof(*start_idx_inc_mod) );
   for ( int i=0; i<start_idx_inc_mod[no_of_inc_modules]; i++ )
   {
      AIMData *adt = new AIMData( file );
      ADinc.push_back( *adt );
      delete adt;
   }
   AMPLinc.resize( start_idx_inc_mod[no_of_inc_modules] );
   for ( int i=0; i<start_idx_inc_mod[no_of_inc_modules]; i++ )
   {
      file.read( (char *) &(AMPLinc[i]), sizeof(AMPLinc[i]) );
   }


   ff = NULL;
}

void AIMsolver::write_binary( fstream &file ) const
{
   file.write( (char *) &veno, sizeof(veno) );
   file.write( (char *) RHS, veno*sizeof(*RHS) );
   file.write( (char *)   J, veno*sizeof(*J)   );


   file.write( (char *) &itno, sizeof(itno) );
   file.write( (char *) &epsilon, sizeof(epsilon) );


   file.write( (char *) &no_of_inc_modules, sizeof(no_of_inc_modules) );
   file.write( (char *)         inc_modno, no_of_inc_modules*sizeof(*inc_modno) );
   file.write( (char *) start_idx_inc_mod, (no_of_inc_modules+1)*sizeof(*start_idx_inc_mod) );
   for ( vector<AIMData>::const_iterator ip = ADinc.begin();
         ip != ADinc.end();
         ip++ )
   {
      ip->write_binary( file );
   }
   for ( int i=0; i<start_idx_inc_mod[no_of_inc_modules]; i++ )
   {
      file.write( (char *) &(AMPLinc[i]), sizeof(AMPLinc[i]) );
   }
}



void AIMsolver::setup()
{
   int m, no;
   POINT p1;
   complex<double> *ca;
   QProgressDialog *PW;

   no_of_inc_modules = 0;

   for ( unsigned int k=0; k<used_modules.size(); k++ )
   {
      ExternalModule *ip = &used_modules[k];

      if ( ip->use_RHS() )
      {
         no_of_inc_modules++;

         try
         {
            if ( ! thread_of_rhs_setup(&*ip) )
               throw( "canceled" );
         }
         catch ( const char *error )
         {
            no_of_inc_modules = 0;
            Xlog << "MODULE( " << k+1 << ": "
                 << ip->name() << " ): "
                 << error << "!" << endl;
            return;
         }
      }
   }

   delete [] RHS;
   delete [] J;
   veno = MAT.start_idx_mod[MAT.no_of_modules];
   RHS = new complex<double>[veno];
   J   = new complex<double>[veno];

   delete [] inc_modno;
   delete [] start_idx_inc_mod;
   inc_modno = new int[ no_of_inc_modules ];
   start_idx_inc_mod = new int[ no_of_inc_modules+1 ];
   start_idx_inc_mod[0] = 0;
   ADinc.clear();
   AMPLinc.clear();

   m = 0;
   for ( unsigned int k=0; k<used_modules.size(); k++ )
   {
      ExternalModule *ip = &used_modules[k];

      if ( ip->use_RHS() )
      {
#ifdef BOUNDARY_CHECK
         if ( m >= no_of_inc_modules )
         {
            clear();
            throw( "AIMsolver::setup index problem" );
         }
#endif
         inc_modno[m] = k;

         ip->RHS_bfkt_inc( &MD, &ca, &no );
         for ( int i=0; i<no; i++ )
         {
            ADinc.push_back( AIMData( MD, ip, ip->no_of_bfkt()+i ) );
            AMPLinc.push_back( ca[i] );
         }
         delete [] ca;

         m++;
         start_idx_inc_mod[m] = start_idx_inc_mod[m-1] + no;
      }
   }
#ifdef BOUNDARY_CHECK
   if ( m != no_of_inc_modules )
      throw( "AIMsolver::setup inc problem" );
#endif


   QString str = QProgressDialog::tr("Calculate right hand side");

   PW = new QProgressDialog( str,
                             QProgressDialog::tr("Cancel"),
                             veno-1,
                             ptrMW,
                             QProgressDialog::tr("progress"),
                             true );
   PW->setProgress(0);

   int e1, idx, kta, jta, ita;
   complex<double> cval, vp, sp, E[3], G[3][3], TJ[3], FTgrfvp,FTgrfdiv;
   AIMfield af( 2*MD.AIMnox-1, 2*MD.AIMnoy-1, 2*MD.AIMnoz-1 );

   for ( int k=0; k<MAT.no_of_modules; k++ )
   {
      ExternalModule *ip = &used_modules[ MAT.modno[k] ];

// J inc
      af.set2zero();
      for ( int m2=0; m2<no_of_inc_modules; m2++ )
      {
         if ( inc_modno[m2] == MAT.modno[k] && ip->has_E_internal() ) continue;   // external only

         for ( int i2=start_idx_inc_mod[m2]; i2<start_idx_inc_mod[m2+1]; i2++ )
            af.add( AMPLinc[i2], ADinc[i2] );
      }
// Einc J
      af.FFT();

      for ( int k3=0; k3<af.noz; k3++ )
      {
         for ( int k2=0; k2<af.noy; k2++ )
         {
            for ( int k1=0; k1<af.nox; k1++ )
            {
/***** zu ungenau?
               grf.FFT_EF_green( G, idx, k1, k2, k3 );
               TJ[0] = af.X[idx];
               TJ[1] = af.Y[idx];
               TJ[2] = af.Z[idx];
               af.X[idx] = G[0][0]*TJ[0] + G[0][1]*TJ[1] + G[0][2]*TJ[2];
               af.Y[idx] = G[1][0]*TJ[0] + G[1][1]*TJ[1] + G[1][2]*TJ[2];
               af.Z[idx] = G[2][0]*TJ[0] + G[2][1]*TJ[1] + G[2][2]*TJ[2];
               af.div[idx] = 0.0;
*****/

               grf.FFT_EP_green( vp, sp, idx, k1, k2, k3 );
               af.X[idx] *= vp;
               af.Y[idx] *= vp;
               af.Z[idx] *= vp;

               af.div[idx] *= sp;
            }
         }
      }

      af.IFFT();

      for ( int i=0; i<ip->no_of_bfkt(); i++ )
      {
         e1 = i + MAT.start_idx_mod[k];

         PW->setProgress( e1 );
         if ( PW->wasCanceled() )
         {
            clear();
            throw( "RHS calculation canceled" );
         }
         ptrAPP->processEvents();

         cval = cNULL;

//
// external contribution

// Einc J
         af.mult_lambda( &vp, &sp, &AD[e1] );
/***** s.o.
         cval -= vp;
*****/
         cval -= -cJ*MD.omega*vp - sp/(cJ*MD.omega);

// Einc
         for ( int ka=0; ka<AD[e1].noz; ka++ )
         {
         kta = ka + AD[e1].z_off;
         p1.z = kta * MD.delta_z + (MD.cpos_z - 0.5*MD.size_z);

         for ( int ja=0; ja<AD[e1].noy; ja++ )
         {
         jta = ja + AD[e1].y_off;
         p1.y = jta * MD.delta_y + (MD.cpos_y - 0.5*MD.size_y);

         for ( int ia=0; ia<AD[e1].nox; ia++ )
         {
         ita = ia + AD[e1].x_off;
         p1.x = ita * MD.delta_x + (MD.cpos_x - 0.5*MD.size_x);

            for ( int k2=0; k2<(int)used_modules.size(); k2++ )
            {
               ExternalModule *ip2 = &used_modules[k2];

               if ( k2 == MAT.modno[k] && ip->has_E_internal() ) continue;   // external only

               if ( ip2->use_RHS() && ip2->has_Einc() )
               {
                  try
                  {
                     ip2->RHS_Einc( &MD, p1.x, p1.y, p1.z, E );

                     idx = AD[e1].index( ia, ja, ka );
                     cval -= (double)AD[e1].lambda_x[ idx ] * E[0];
                     cval -= (double)AD[e1].lambda_y[ idx ] * E[1];
                     cval -= (double)AD[e1].lambda_z[ idx ] * E[2];
                  }
                  catch( const char *str )
                  {
                     PW->cancel();
                     throw(str);
                  }
               }
            }

         }
         }
         }

//
// internal contribution
         if ( ip->use_RHS() )
         {
            if ( ip->has_E_internal() )
            {
               try
               {
                  cval -= ip->RHS_E_internal( &MD,
                                              e1 - MAT.start_idx_mod[k] );
               }
               catch( const char *str )
               {
                  PW->cancel();
                  throw(str);
               }
            }
            else
            {
               Xlog << "WARNING: module \"" << ip->name() << "\""
                    << " has no E_internal!" << endl;
            }
         }

         RHS[e1] = cval;
         J[e1] = cNULL;
      }
   }

   PW->cancel();
   delete PW;


   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      ip->RHS_E_cleanup( &MD );
   }

}


void AIMsolver::precon( complex<double> *in, complex<double> *out )
{
   int k;
   complex<double> sum;

// L
   for ( int i=0; i<MAT.L->nofrows; i++ )
   {
      sum = cNULL;
      for ( int j=MAT.L->first_ele[i]; j<MAT.L->first_ele[i+1]-1; j++ )
         sum += (complex<double>)MAT.L->ele[j] * out[ MAT.L->colid[j] ];

      k = MAT.L->first_ele[i+1] - 1;
#ifdef BOUNDARY_CHECK
      if ( MAT.L->colid[k] != i )
      {
         cerr << "Error in precon(): line "
              << __LINE__ << "!" << endl;
         exit(1);
      }
#endif
      out[i] = ( in[i] - sum ) / (complex<double>)MAT.L->ele[k];
   }

// LT
   for ( int i=MAT.L->nofrows-1; i>-1; i-- )
   {
      k = MAT.L->first_ele[i+1] - 1;
#ifdef BOUNDARY_CHECK
      if ( MAT.L->colid[k] != i )
      {
         cerr << "Error in precon(): line "
              << __LINE__ << "!" << endl;
         exit(1);
      }
#endif
      sum = out[i] = out[i] / (complex<double>)MAT.L->ele[k];

      for ( int j=MAT.L->first_ele[i]; j<k; j++ )
      {
         out[ MAT.L->colid[j] ] -= (complex<double>)MAT.L->ele[j] * sum;
      }
   }

}


void AIMsolver::matvecmult( AIMfield *af,
                            complex<double> *x, complex<double> *ax )
{
   int m1, m2, k1, k2, k3, idx;
   complex<double> vp, phi;

   MAT.Z->vecmult( x, ax );

   af->set2zero();
   for ( m2=0; m2<(int)AD.size(); m2++ )
      af->add( x[m2], AD[m2] );

   af->FFT();

   for ( k3=0; k3<af->noz; k3++ )
   {
      for ( k2=0; k2<af->noy; k2++ )
      {
         for ( k1=0; k1<af->nox; k1++ )
         {
            grf.FFT_EP_green( vp, phi, idx, k1, k2, k3 );
            af->X[idx]  *= vp;
            af->Y[idx]  *= vp;
            af->Z[idx]  *= vp;
            af->div[idx] *= phi;
         }
      }
   }

   af->IFFT();

   for ( m1=0; m1<(int)AD.size(); m1++ )
   {
      af->mult_lambda( &vp, &phi, &(AD[m1]) );
      ax[m1] -= cJ*MD.omega * vp + phi / (cJ*MD.omega);
   }

}


void AIMsolver::solve( double &eps, int &max_itno )
{

   int i, iter, ui;
   double epsilon, norm_rhs;
   complex<double> *r, *p, *x, *z, *ap, ab, skp;
   AIMfield aim_field( 2*MD.AIMnox-1, 2*MD.AIMnoy-1, 2*MD.AIMnoz-1 );
   QProgressDialog *PW;
   QString str;

   p = new complex<double>[veno];
   r = new complex<double>[veno];
   z = new complex<double>[veno];
   ap = new complex<double>[veno];

   ui = 0;
   str = QProgressDialog::tr("Solving");
   PW = new QProgressDialog( str,
                             QProgressDialog::tr("Cancel"),
                             0,
                             ptrMW,
                             QProgressDialog::tr("progress"),
                             true );
   PW->setProgress(ui);

// init
   x = J;
   matvecmult( &aim_field, x, ap );
#pragma omp parallel for
   for ( i=0; i<veno; i++ ) r[i] = RHS[i] - ap[i];

   precon( r, z );

   for ( i=0; i<veno; i++ ) p[i] = z[i];

   norm_rhs = 0.0;
#pragma omp parallel for reduction(+:norm_rhs)
   for ( i=0; i<veno; i++ ) norm_rhs += abs(RHS[i]*conj(RHS[i]));

// iterate
   iter = 0;
   epsilon = 0.0;
#pragma omp parallel for reduction(+:epsilon)
   for ( i=0; i<veno; i++ ) epsilon += abs(r[i]*conj(r[i]));
   epsilon = sqrt( epsilon/norm_rhs );

   skp = cNULL;
   for ( i=0; i<veno; i++ ) skp += r[i] * z[i];

   while ( epsilon > eps && iter < max_itno )
   {
      ui += 10;
      PW->setProgress(ui);
      if ( PW->wasCanceled() )
      {
         max_itno = iter;
         eps = epsilon;

         delete [] ap;
         ap = NULL;
         delete [] z;
         z = NULL;
         delete [] r;
         r = NULL;
         delete [] p;
         p = NULL;

         throw( "Solving canceled" );
      }

      matvecmult( &aim_field, p, ap );
      ab = cNULL;
      for ( i=0; i<veno; i++ ) ab += p[i] * ap[i];
      ab = skp/ab;

#pragma omp parallel for firstprivate(ab)
      for ( i=0; i<veno; i++ )
      {
         x[i] += ab * p[i];
         r[i] -= ab * ap[i];
      }

      precon( r, z );
      ab = 1.0 / skp;
      skp = cNULL;
      for ( i=0; i<veno; i++ ) skp += r[i] * z[i];
      ab *= skp;

#pragma omp parallel for firstprivate(ab)
      for ( i=0; i<veno; i++ ) p[i] = z[i] + ab * p[i];

      iter++;
      epsilon = 0.0;
#pragma omp parallel for reduction(+:epsilon)
      for ( i=0; i<veno; i++ ) epsilon += abs(r[i]*conj(r[i]));
      epsilon = sqrt( epsilon/norm_rhs );
   }

   max_itno = iter;
   eps = epsilon;

   PW->cancel();
   delete PW;

   delete [] ap;
   delete [] z;
   delete [] r;
   delete [] p;
}



void AIMsolver::clear()
{
   delete [] J;
   delete [] RHS;

   veno = 0;
   RHS = J = NULL;

   itno = 0;
   epsilon = 0.0;

   no_of_inc_modules = 0;
   delete [] inc_modno;
   inc_modno = NULL;
   delete [] start_idx_inc_mod;
   start_idx_inc_mod = NULL;
   ADinc.clear();
   AMPLinc.clear();

   delete ff;
   ff = NULL;
}


void AIMsolver::extract()
{
   POINT p1;
   FIELDx field( MD );
   AIMfield af( 2*MD.AIMnox-1, 2*MD.AIMnoy-1, 2*MD.AIMnoz-1 );
   int m, m2, k1, k2, k3, i1, i2, i3, aidx, fidx;
   complex<double> E[3], G[3][3], TJ[3];

   m = 0;
   for ( unsigned int k=0; k<used_modules.size(); k++ )
   {
      ExternalModule *ip = &used_modules[k];
      m++;

      if ( ! ip->use_extract() ) continue;

      field.set2zero();


// Einc
      for ( unsigned int kk=0; kk<used_modules.size(); kk++ )
      {
         ExternalModule *ip2 = &used_modules[kk];

         if ( kk == k || (! ip2->use_RHS())                // external only
                      || (! ip2->has_Einc()) ) continue;   // no Einc

         for ( int ka=0; ka<MD.AIMnoz; ka++ )
         {
         p1.z = ka * MD.delta_z + (MD.cpos_z - 0.5*MD.size_z);

         for ( int ja=0; ja<MD.AIMnoy; ja++ )
         {
         p1.y = ja * MD.delta_y + (MD.cpos_y - 0.5*MD.size_y);

         for ( int ia=0; ia<MD.AIMnox; ia++ )
         {
         p1.x = ia * MD.delta_x + (MD.cpos_x - 0.5*MD.size_x);

            ip2->RHS_Einc( &MD, p1.x, p1.y, p1.z, E );

            fidx = ia + field.nox * ja + field.nox*field.noy * ka;
            field.Einc[ fidx ].x += E[0];
            field.Einc[ fidx ].y += E[1];
            field.Einc[ fidx ].z += E[2];
         }
         }
         }

      }


// J scatt
      af.set2zero();
      for ( m2=0; m2<(int)AD.size(); m2++ )
      {
         if ( m2 <  ip->matrix_offset() ||
              m2 >= ip->matrix_offset() + ip->no_of_bfkt() )
            af.add( J[m2], AD[m2] );
      }

      for ( i3=0; i3<field.noz; i3++ )
      {
         for ( i2=0; i2<field.noy; i2++ )
         {
            for ( i1=0; i1<field.nox; i1++ )
            {
               fidx = i1 + field.nox * i2 + field.nox*field.noy * i3;
               aidx = i1 + af.nox * i2 + af.nox*af.noy * i3;
               field.J[ fidx ].x = af.X[ aidx ];
               field.J[ fidx ].y = af.Y[ aidx ];
               field.J[ fidx ].z = af.Z[ aidx ];
            }
         }
      }


      af.FFT();

      for ( k3=0; k3<af.noz; k3++ )
      {
         for ( k2=0; k2<af.noy; k2++ )
         {
            for ( k1=0; k1<af.nox; k1++ )
            {
               grf.FFT_EF_green( G, aidx, k1, k2, k3 );
               TJ[0] = af.X[aidx];
               TJ[1] = af.Y[aidx];
               TJ[2] = af.Z[aidx];
               af.X[aidx] = G[0][0]*TJ[0] + G[0][1]*TJ[1] + G[0][2]*TJ[2];
               af.Y[aidx] = G[1][0]*TJ[0] + G[1][1]*TJ[1] + G[1][2]*TJ[2];
               af.Z[aidx] = G[2][0]*TJ[0] + G[2][1]*TJ[1] + G[2][2]*TJ[2];
            }
         }
      }

      af.IFFT();

      for ( i3=0; i3<field.noz; i3++ )
      {
         for ( i2=0; i2<field.noy; i2++ )
         {
            for ( i1=0; i1<field.nox; i1++ )
            {
               fidx = i1 + field.nox * i2 + field.nox*field.noy * i3;
               aidx = i1 + af.nox * i2 + af.nox*af.noy * i3;
               field.Escatt[ fidx ].x = af.X[ aidx ];
               field.Escatt[ fidx ].y = af.Y[ aidx ];
               field.Escatt[ fidx ].z = af.Z[ aidx ];
            }
         }
      }


// J inc
      af.set2zero();
      for ( m2=0; m2<no_of_inc_modules; m2++ )
      {
         if ( inc_modno[m2] != (int)k )
            for ( i2=start_idx_inc_mod[m2]; i2<start_idx_inc_mod[m2+1]; i2++ )
            {
               af.add( AMPLinc[i2], ADinc[i2] );
            }
      }

      for ( i3=0; i3<field.noz; i3++ )
      {
         for ( i2=0; i2<field.noy; i2++ )
         {
            for ( i1=0; i1<field.nox; i1++ )
            {
               fidx = i1 + field.nox * i2 + field.nox*field.noy * i3;
               aidx = i1 + af.nox * i2 + af.nox*af.noy * i3;
               field.J[ fidx ].x += af.X[ aidx ];
               field.J[ fidx ].y += af.Y[ aidx ];
               field.J[ fidx ].z += af.Z[ aidx ];
            }
         }
      }

// Einc J
      af.FFT();

      for ( k3=0; k3<af.noz; k3++ )
      {
         for ( k2=0; k2<af.noy; k2++ )
         {
            for ( k1=0; k1<af.nox; k1++ )
            {
               grf.FFT_EF_green( G, aidx, k1, k2, k3 );
               TJ[0] = af.X[aidx];
               TJ[1] = af.Y[aidx];
               TJ[2] = af.Z[aidx];
               af.X[aidx] = G[0][0]*TJ[0] + G[0][1]*TJ[1] + G[0][2]*TJ[2];
               af.Y[aidx] = G[1][0]*TJ[0] + G[1][1]*TJ[1] + G[1][2]*TJ[2];
               af.Z[aidx] = G[2][0]*TJ[0] + G[2][1]*TJ[1] + G[2][2]*TJ[2];
            }
         }
      }

      af.IFFT();

      for ( i3=0; i3<field.noz; i3++ )
      {
         for ( i2=0; i2<field.noy; i2++ )
         {
            for ( i1=0; i1<field.nox; i1++ )
            {
               fidx = i1 + field.nox * i2 + field.nox*field.noy * i3;
               aidx = i1 + af.nox * i2 + af.nox*af.noy * i3;
               field.Einc[ fidx ].x += af.X[ aidx ];
               field.Einc[ fidx ].y += af.Y[ aidx ];
               field.Einc[ fidx ].z += af.Z[ aidx ];
            }
         }
      }


      try
      {
         if ( ! thread_of_extract_setup( &*ip, &field ) )
            throw( "canceled" );
      }
      catch ( const char *error )
      {
         Xlog << "MODULE( " << m << ": "
              << ip->name() << " ): "
              << error << "!" << endl;
         throw( "extraction canceled" );
      }

   }


//
// GUI extract functions
   field.set2zero();


// Einc
   for ( unsigned int kk=0; kk<used_modules.size(); kk++ )
   {
      ExternalModule *ip2 = &used_modules[kk];

      if ( (! ip2->use_RHS()) || (! ip2->has_Einc()) ) continue;

      for ( int ka=0; ka<MD.AIMnoz; ka++ )
      {
      p1.z = ka * MD.delta_z + (MD.cpos_z - 0.5*MD.size_z);

      for ( int ja=0; ja<MD.AIMnoy; ja++ )
      {
      p1.y = ja * MD.delta_y + (MD.cpos_y - 0.5*MD.size_y);

      for ( int ia=0; ia<MD.AIMnox; ia++ )
      {
      p1.x = ia * MD.delta_x + (MD.cpos_x - 0.5*MD.size_x);

         ip2->RHS_Einc( &MD, p1.x, p1.y, p1.z, E );

         fidx = ia + field.nox * ja + field.nox*field.noy * ka;
         field.Einc[ fidx ].x += E[0];
         field.Einc[ fidx ].y += E[1];
         field.Einc[ fidx ].z += E[2];
      }
      }
      }

   }


// J scatt
   af.set2zero();
   for ( m2=0; m2<(int)AD.size(); m2++ )
   {
      af.add( J[m2], AD[m2] );
   }

   for ( i3=0; i3<field.noz; i3++ )
   {
      for ( i2=0; i2<field.noy; i2++ )
      {
         for ( i1=0; i1<field.nox; i1++ )
         {
            fidx = i1 + field.nox * i2 + field.nox*field.noy * i3;
            aidx = i1 + af.nox * i2 + af.nox*af.noy * i3;
            field.J[ fidx ].x += af.X[ aidx ];
            field.J[ fidx ].y += af.Y[ aidx ];
            field.J[ fidx ].z += af.Z[ aidx ];
         }
      }
   }

   af.FFT();

   for ( k3=0; k3<af.noz; k3++ )
   {
      for ( k2=0; k2<af.noy; k2++ )
      {
         for ( k1=0; k1<af.nox; k1++ )
         {
            grf.FFT_EF_green( G, aidx, k1, k2, k3 );
            TJ[0] = af.X[aidx];
            TJ[1] = af.Y[aidx];
            TJ[2] = af.Z[aidx];
            af.X[aidx] = G[0][0]*TJ[0] + G[0][1]*TJ[1] + G[0][2]*TJ[2];
            af.Y[aidx] = G[1][0]*TJ[0] + G[1][1]*TJ[1] + G[1][2]*TJ[2];
            af.Z[aidx] = G[2][0]*TJ[0] + G[2][1]*TJ[1] + G[2][2]*TJ[2];
         }
      }
   }

   af.IFFT();

   for ( i3=0; i3<field.noz; i3++ )
   {
      for ( i2=0; i2<field.noy; i2++ )
      {
         for ( i1=0; i1<field.nox; i1++ )
         {
            fidx = i1 + field.nox * i2 + field.nox*field.noy * i3;
            aidx = i1 + af.nox * i2 + af.nox*af.noy * i3;
            field.Escatt[ fidx ].x += af.X[ aidx ];
            field.Escatt[ fidx ].y += af.Y[ aidx ];
            field.Escatt[ fidx ].z += af.Z[ aidx ];
         }
      }
   }


// J inc
   af.set2zero();
   for ( m2=0; m2<no_of_inc_modules; m2++ )
   {
      for ( i2=start_idx_inc_mod[m2]; i2<start_idx_inc_mod[m2+1]; i2++ )
      {
         af.add( AMPLinc[i2], ADinc[i2] );
      }
   }

   for ( i3=0; i3<field.noz; i3++ )
   {
      for ( i2=0; i2<field.noy; i2++ )
      {
         for ( i1=0; i1<field.nox; i1++ )
         {
            fidx = i1 + field.nox * i2 + field.nox*field.noy * i3;
            aidx = i1 + af.nox * i2 + af.nox*af.noy * i3;
            field.J[ fidx ].x += af.X[ aidx ];
            field.J[ fidx ].y += af.Y[ aidx ];
            field.J[ fidx ].z += af.Z[ aidx ];
         }
      }
   }

// Einc J
   af.FFT();

   for ( k3=0; k3<af.noz; k3++ )
   {
      for ( k2=0; k2<af.noy; k2++ )
      {
         for ( k1=0; k1<af.nox; k1++ )
         {
            grf.FFT_EF_green( G, aidx, k1, k2, k3 );
            TJ[0] = af.X[aidx];
            TJ[1] = af.Y[aidx];
            TJ[2] = af.Z[aidx];
            af.X[aidx] = G[0][0]*TJ[0] + G[0][1]*TJ[1] + G[0][2]*TJ[2];
            af.Y[aidx] = G[1][0]*TJ[0] + G[1][1]*TJ[1] + G[1][2]*TJ[2];
            af.Z[aidx] = G[2][0]*TJ[0] + G[2][1]*TJ[1] + G[2][2]*TJ[2];
         }
      }
   }

   af.IFFT();

   for ( i3=0; i3<field.noz; i3++ )
   {
      for ( i2=0; i2<field.noy; i2++ )
      {
         for ( i1=0; i1<field.nox; i1++ )
         {
            fidx = i1 + field.nox * i2 + field.nox*field.noy * i3;
            aidx = i1 + af.nox * i2 + af.nox*af.noy * i3;
            field.Einc[ fidx ].x += af.X[ aidx ];
            field.Einc[ fidx ].y += af.Y[ aidx ];
            field.Einc[ fidx ].z += af.Z[ aidx ];
         }
      }
   }



   if ( MD.coeff )
   {
      complex<double> *CV;
      int lstart, lend, offset;

      Xlog << endl;
      Xlog << "output of ";
      switch ( MD.coeff_vec )
      {
         case  0: { CV = J; Xlog << "solution vector"; break; }
         case  1: { CV = RHS; Xlog << "RHS vector"; break; }
         default: { throw( "unknown selection" ); }
      }
      Xlog << endl;

      Xlog << " numbering scheme: ";
      if ( MD.coeff_numbering == 0 )
      {
         Xlog << "global" << endl;
         lstart = MD.coeff_from;
         if ( lstart < 0 ) lstart = 0;
         if ( lstart > veno ) lstart = veno;
         lend = MD.coeff_to + 1;
         if ( lend < 0 ) lend = 0;
         if ( lend > veno ) lend = veno;
         offset = 0;
      }
      else if ( MD.coeff_numbering > 0 && MD.coeff_numbering <= (int)used_modules.size() )
      {
         Xlog << used_modules[ MD.coeff_numbering-1 ].name()
              << "(" << MD.coeff_numbering << ")" << endl;

         lstart = MD.coeff_from;
         if ( lstart < 0 )
            lstart = 0;
         if ( lstart > used_modules[ MD.coeff_numbering-1 ].no_of_bfkt() )
            lstart = used_modules[ MD.coeff_numbering-1 ].no_of_bfkt();

         lend = MD.coeff_to + 1;
         if ( lend < 0 )
            lend = 0;
         if ( lend > used_modules[ MD.coeff_numbering-1 ].no_of_bfkt() )
            lend = used_modules[ MD.coeff_numbering-1 ].no_of_bfkt();

         offset = used_modules[ MD.coeff_numbering-1 ].matrix_offset();
      }
      else
         throw( "unknown module number" );

      for ( int i=lstart; i<lend; i++ )
      {
         Xlog << " " << i << ": " << CV[ i + offset ] << endl;
      }
   }


   if ( MD.near_field )
   {
      Xlog << endl;
      Xlog << "near field probe "
           << "at ( " << MD.nfp_x << "; "
                      << MD.nfp_y << "; "
                      << MD.nfp_z << " ) m" << endl;
      Xlog << endl;
      field.E_inc( MD.nfp_x, MD.nfp_y, MD.nfp_z, E );
      Xlog << " E_inc_x: ( " << E[0].real() << "; " << E[0].imag() << " ) V/m" << endl;
      Xlog << " E_inc_y: ( " << E[1].real() << "; " << E[1].imag() << " ) V/m" << endl;
      Xlog << " E_inc_z: ( " << E[2].real() << "; " << E[2].imag() << " ) V/m" << endl;
      Xlog << " abs(E_inc): " << sqrt( abs( E[0]*conj(E[0])
                                            + E[1]*conj(E[1])
                                            + E[2]*conj(E[2]) ) ) << " V/m" << endl;
      Xlog << endl;
      field.E_scatt( MD.nfp_x, MD.nfp_y, MD.nfp_z, E );
      Xlog << " E_scatt_x: ( " << E[0].real() << "; " << E[0].imag() << " ) V/m" << endl;
      Xlog << " E_scatt_y: ( " << E[1].real() << "; " << E[1].imag() << " ) V/m" << endl;
      Xlog << " E_scatt_z: ( " << E[2].real() << "; " << E[2].imag() << " ) V/m" << endl;
      Xlog << " abs(E_scatt): " << sqrt( abs( E[0]*conj(E[0])
                                            + E[1]*conj(E[1])
                                            + E[2]*conj(E[2]) ) ) << " V/m" << endl;
      Xlog << endl;
      field.E_total( MD.nfp_x, MD.nfp_y, MD.nfp_z, E );
      Xlog << " E_total_x: ( " << E[0].real() << "; " << E[0].imag() << " ) V/m" << endl;
      Xlog << " E_total_y: ( " << E[1].real() << "; " << E[1].imag() << " ) V/m" << endl;
      Xlog << " E_total_z: ( " << E[2].real() << "; " << E[2].imag() << " ) V/m" << endl;
      Xlog << " abs(E_total): " << sqrt( abs( E[0]*conj(E[0])
                                            + E[1]*conj(E[1])
                                            + E[2]*conj(E[2]) ) ) << " V/m" << endl;
      Xlog << endl;



      field.E_Jtotal_sum( MD.nfp_x, MD.nfp_y, MD.nfp_z, E );
      Xlog << " E_Jtotal_x: ( " << E[0].real() << "; " << E[0].imag() << " ) V/m" << endl;
      Xlog << " E_Jtotal_y: ( " << E[1].real() << "; " << E[1].imag() << " ) V/m" << endl;
      Xlog << " E_Jtotal_z: ( " << E[2].real() << "; " << E[2].imag() << " ) V/m" << endl;
      Xlog << " abs(E_Jtotal): " << sqrt( abs( E[0]*conj(E[0])
                                               + E[1]*conj(E[1])
                                               + E[2]*conj(E[2]) ) ) << " V/m" << endl;
   }

   delete ff;
   ff = NULL;
   if ( MD.far_field )
   {
      ff = new FARFIELD( &field, MD.pad_x, MD.pad_y, MD.pad_z );

      if ( MD.radiated_pwr )
      {
         double pwr = ff->rad_power( MD.no_theta, MD.no_phi, MD.itpl_deg );

         Xlog << endl;
         Xlog << " radiated power: " << pwr << " W" << endl;
      }

   }

}
