//
// Class for far-field
// (c) 16.4.2007 Ingo D. Rullhusen
//

#include <fftw3.h>

#include <iostream>

#include <qprogressdialog.h>

#include "natconst.h"
#include "legendre.h"
#include "global_vars.h"

#include "farfield.h"


FARFIELD::FARFIELD( FIELDx *f, int pad_x, int pad_y, int pad_z )
{
   nox = ( f->nox > 1 ? f->nox + pad_x : 1 );
   noy = ( f->noy > 1 ? f->noy + pad_y : 1 );
   noz = ( f->noz > 1 ? f->noz + pad_z : 1 );

   FFT_Jx = new complex<double>[ nox*noy*noz ];
   FFT_Jy = new complex<double>[ nox*noy*noz ];
   FFT_Jz = new complex<double>[ nox*noy*noz ];

   delta_kx = ( f->nox > 1 ? 2.0*M_PI / ( nox * f->delta_x ) : 0.0 );
   delta_ky = ( f->noy > 1 ? 2.0*M_PI / ( noy * f->delta_y ) : 0.0 );
   delta_kz = ( f->noz > 1 ? 2.0*M_PI / ( noz * f->delta_z ) : 0.0 );

   lpos_x = f->lpos_x;
   lpos_y = f->lpos_y;
   lpos_z = f->lpos_z;

   k0 = f->k0;

   facE = complex<double>( 0.0, -f->k0*ZF0 / (4.0*M_PI) );


   fftw_plan px, py, pz;
#pragma omp critical (FFTW)
{
   px = fftw_plan_dft_3d( noz, noy, nox,
                          (fftw_complex*)FFT_Jx, (fftw_complex*)FFT_Jx,
                          FFTW_BACKWARD, FFTW_ESTIMATE
                        );
   py = fftw_plan_dft_3d( noz, noy, nox,
                          (fftw_complex*)FFT_Jy, (fftw_complex*)FFT_Jy,
                          FFTW_BACKWARD, FFTW_ESTIMATE
                        );
   pz = fftw_plan_dft_3d( noz, noy, nox,
                          (fftw_complex*)FFT_Jz, (fftw_complex*)FFT_Jz,
                          FFTW_BACKWARD, FFTW_ESTIMATE
                        );
}  // end:pragma omp critical (FFTW)

   for ( int i=0; i<nox*noy*noz; i++ ) FFT_Jx[i] = FFT_Jy[i] = FFT_Jz[i] = complex<double>(0.0,0.0);

   for ( int k=0; k<f->noz; k++ )
   {
      for ( int j=0; j<f->noy; j++ )
      {
         for ( int i=0; i<f->nox; i++ )
         {
            FFT_Jx[ i + j*nox + k*nox*noy ] = f->J[ i + j*f->nox + k*f->nox*f->noy ].x;
            FFT_Jy[ i + j*nox + k*nox*noy ] = f->J[ i + j*f->nox + k*f->nox*f->noy ].y;
            FFT_Jz[ i + j*nox + k*nox*noy ] = f->J[ i + j*f->nox + k*f->nox*f->noy ].z;
         }
      }
   }

   fftw_execute(px);
   fftw_execute(py);
   fftw_execute(pz);

#pragma omp critical (FFTW)
{
   fftw_destroy_plan(px);
   fftw_destroy_plan(py);
   fftw_destroy_plan(pz);
}  // end:pragma omp critical (FFTW)
}


FARFIELD::~FARFIELD()
{
   delete [] FFT_Jx;
   delete [] FFT_Jy;
   delete [] FFT_Jz;
}


void FARFIELD::FT_J( complex<double> ftj[3], double kx, double ky, double kz, int pktno )
{
   int nx, ny, nz, i, j, k, m, idx, idy, idz, itp;
   int ix[pktno], iy[pktno], iz[pktno];
   double pm, pkx, pky, pkz, ipfx, ipfy, ipfz;

   nx = ( nox==1 ? nox : pktno );
   ny = ( noy==1 ? noy : pktno );
   nz = ( noz==1 ? noz : pktno );

   ix[0] = ( nox > 1 ? (int)ceil( kx/delta_kx ) - nx/2 : 0 );
   if ( abs(ix[0]) > nox/2 ) throw( "far-field x aliasing" );
   if ( ix[0] < 0 ) ix[0] += nox;

   for ( i=1; i<nx; i++ )
   {
      ix[i] = ix[i-1] + 1;
      if ( ix[i] >= nox ) ix[i] -= nox;
      if ( ix[i] >= nox ) throw( "far-field x aliasing" );
   }


   iy[0] = ( noy > 1 ? (int)ceil( ky/delta_ky ) - ny/2 : 0 );
   if ( abs(iy[0]) > noy/2 ) throw( "far-field y aliasing" );
   if ( iy[0] < 0 ) iy[0] += noy;

   for ( i=1; i<ny; i++ )
   {
      iy[i] = iy[i-1] + 1;
      if ( iy[i] >= noy ) iy[i] -= noy;
      if ( iy[i] >= noy ) throw( "far-field y aliasing" );
   }


   iz[0] = ( noz > 1 ? (int)ceil( kz/delta_kz ) - nz/2 : 0 );
   if ( abs(iz[0]) > noz/2 ) throw( "far-field z aliasing" );
   if ( iz[0] < 0 ) iz[0] += noz;

   for ( i=1; i<nz; i++ )
   {
      iz[i] = iz[i-1] + 1;
      if ( iz[i] >= noz ) iz[i] -= noz;
      if ( iz[i] >= noz ) throw( "far-field z aliasing" );
   }


   int no = nx*ny*nz;
   int *index = new int[no];
   double *ipf = new double[no];
   complex<double> *lenfac = new complex<double>[no];

   itp = 0;
   for ( k=0; k<nz; k++ )
   {
      idz = nox*noy * iz[k];
      ipfz = 1.0;
      pkz = delta_kz * ( iz[k]>noz/2 ? iz[k]-noz : iz[k] );
      for ( m=0; m<nz; m++ )
      {
         pm = delta_kz * ( iz[m]>noz/2 ? iz[m]-noz : iz[m] );
         if ( m != k ) ipfz *= ( kz - pm ) / ( pkz - pm );
      }

      for ( j=0; j<ny; j++ )
      {
         idy = nox * iy[j];
         ipfy = 1.0;
         pky = delta_ky * ( iy[j]>noy/2 ? iy[j]-noy : iy[j] );
         for ( m=0; m<ny; m++ )
         {
            pm = delta_ky * ( iy[m]>noy/2 ? iy[m]-noy : iy[m] );
            if ( m != j ) ipfy *= ( ky - pm ) / ( pky - pm );
         }

         for ( i=0; i<nx; i++ )
         {
            idx = ix[i];
            ipfx = 1.0;
            pkx = delta_kx * ( ix[i]>nox/2 ? ix[i]-nox : ix[i] );
            for ( m=0; m<nx; m++ )
            {
               pm = delta_kx * ( ix[m]>nox/2 ? ix[m]-nox : ix[m] );
               if ( m != i ) ipfx *= ( kx - pm ) / ( pkx - pm );
            }

            index[itp] = idx + idy + idz;
            ipf[itp] = ipfx * ipfy * ipfz;

            lenfac[itp] = exp( complex<double>( 0.0, pkx*lpos_x+pky*lpos_y+pkz*lpos_z ) );

            itp++;
         }
      }
   }


   ftj[0] = ftj[1] = ftj[2] = complex<double>(0.0,0.0);

   for ( i=0; i<no; i++ )
   {
      ftj[0] += lenfac[i] * FFT_Jx[ index[i] ] * ipf[i];
      ftj[1] += lenfac[i] * FFT_Jy[ index[i] ] * ipf[i];
      ftj[2] += lenfac[i] * FFT_Jz[ index[i] ] * ipf[i];
   }


   delete [] lenfac;

   delete [] index;
   delete [] ipf;
}

void FARFIELD::E_ff( complex<double> E[3], double x, double y, double z, int itpl )
{
   double r, rv[3];
   complex<double> ftj[3], c;

   r = sqrt( x*x + y*y + z*z );
   rv[0] = x/r;
   rv[1] = y/r;
   rv[2] = z/r;

   FT_J( ftj, k0*rv[0], k0*rv[1], k0*rv[2], itpl+1 );

   c = rv[0]*ftj[0] + rv[1]*ftj[1] + rv[2]*ftj[2];

   ftj[0] -= c*rv[0];
   ftj[1] -= c*rv[1];
   ftj[2] -= c*rv[2];

   E[0] = facE * ftj[0];
   E[1] = facE * ftj[1];
   E[2] = facE * ftj[2];
}



double FARFIELD::rad_power( int notheta, int nophi, int itpl )
{
   int i, j;
   double pwr;
   double *costh, *sinth, *cosphi, *sinphi, *tw;
   complex<double> E[3];


   costh = new double[notheta];
   sinth = new double[notheta];
   i = legendre_zero( costh, notheta, 0.001 );
   if ( i )
   {
      throw( "cannot find Legendre zeros" );
   }
   for ( i=0; i<notheta; i++ ) sinth[i] = sqrt( 1.0 - costh[i]*costh[i] );


   cosphi = new double[nophi];
   sinphi = new double[nophi];
   for ( i=0; i<nophi; i++ )  
   {
      cosphi[i] = cos( i*( 2.0*M_PI / nophi ) );
      sinphi[i] = sin( i*( 2.0*M_PI / nophi ) );
   }


   tw = new double[notheta];
   i = legendre_weights( tw, costh, notheta );
   if ( i )
   {
      throw( "cannot find Legendre weights" );
   }



   QProgressDialog *PW;

   QString str = QProgressDialog::tr("integrate far-field...");

   PW = new QProgressDialog( str,
                             QProgressDialog::tr("Cancel"),
                             notheta,
                             ptrMW,
                             QProgressDialog::tr("progress"),
                             true );
   PW->setProgress(0);


   pwr = 0.0;

   for ( j=0; j<notheta; j++ )
   {

      PW->setProgress( j+1 );
      if ( PW->wasCanceled() )
      {
         throw( "far-field integration canceled" );
      }
      ptrAPP->processEvents();


      try
      {

#pragma omp parallel for reduction(+:pwr) private(i,E)
      for ( i=0; i<nophi; i++ )
      {

         E_ff( E, sinth[j]*cosphi[i], sinth[j]*sinphi[i], costh[j], itpl );

         pwr += tw[j] * 0.5 * abs(E[0]*conj(E[0]) + E[1]*conj(E[1]) + E[2]*conj(E[2])) / ZF0;

      }

      }
      catch( const char *e )
      {
         Xlog << "problem with far-field: " << e << endl;
         PW->cancel();
      }
   }

   pwr *= 2.0*M_PI / nophi;


   delete PW;

   delete [] costh;
   delete [] sinth;
   delete [] cosphi;
   delete [] sinphi;
   delete [] tw;

   return(pwr);
}

