//
// Class for FIELDx grid
// (c) 16.4.2007 Ingo D. Rullhusen
//


#include <iostream>
#include "natconst.h"
#include "auxiliary.h"

#include <float.h>

#include "fieldx.h"

#include "global_vars.h"


FIELDx::FIELDx( MainDataX MD )
{
   nox = MD.AIMnox;
   noy = MD.AIMnoy;
   noz = MD.AIMnoz;

   int noxyz = nox*noy*noz;

   Einc = new cVECTOR[ noxyz ];
   Escatt = new cVECTOR[ noxyz ];
   J = new cVECTOR[ noxyz ];

   delta_x = MD.delta_x;
   delta_y = MD.delta_y;
   delta_z = MD.delta_z;

   lpos_x = MD.cpos_x - 0.5*MD.size_x;
   lpos_y = MD.cpos_y - 0.5*MD.size_y;
   lpos_z = MD.cpos_z - 0.5*MD.size_z;


   k0 = MD.k0;

   set2zero();
}

FIELDx::~FIELDx()
{
   delete [] Einc;
   delete [] Escatt;
   delete [] J;
}


void FIELDx::set2zero()
{
   const complex<double> cNULL(0.0,0.0);
   int noxyz = nox*noy*noz;

   for ( int i=0; i<noxyz; i++ )
   {
      Einc[i].x = Einc[i].y = Einc[i].z = cNULL;
      Escatt[i].x = Escatt[i].y = Escatt[i].z = cNULL;
      J[i].x = J[i].y = J[i].z = cNULL;
   }
}


void FIELDx::itpl( double x, double y, double z,
                   int &no, int **index, double **ipf )
{
   const int maxpkt = 3;
   int nx, ny, nz, i, j, k, m, idx, idy, idz, itp;
   int ix[maxpkt], iy[maxpkt], iz[maxpkt];
   double pm, pk, ipfx, ipfy, ipfz;


   ix[0] = ( nox > 1 ? (int)ceil( (x-lpos_x)/delta_x - 0.01*delta_x ) - maxpkt/2 : 0 );
   ix[0] = ( ix[0] < 0 ? 0 : ix[0] );
   ix[0] = ( ix[0] >= nox ? nox-1 : ix[0] );

   nx = 1;

   for ( i=1; i<maxpkt; i++ )
   {
      ix[i] = ix[i-1] + 1;
      if ( ix[i] < nox )
         nx++;
      else
         break;
   }

   iy[0] = ( noy > 1 ? (int)ceil( (y-lpos_y)/delta_y - 0.01*delta_y ) - maxpkt/2 : 0 );
   iy[0] = ( iy[0] < 0 ? 0 : iy[0] );
   iy[0] = ( iy[0] >= noy ? noy-1 : iy[0] );

   ny = 1;

   for ( i=1; i<maxpkt; i++ )
   {
      iy[i] = iy[i-1] + 1;
      if ( iy[i] < noy )
         ny++;
      else
         break;
   }

   iz[0] = ( noz > 1 ? (int)ceil( (z-lpos_z)/delta_z - 0.01*delta_z ) - maxpkt/2 : 0 );
   iz[0] = ( iz[0] < 0 ? 0 : iz[0] );
   iz[0] = ( iz[0] >= noz ? noz-1 : iz[0] );

   nz = 1;

   for ( i=1; i<maxpkt; i++ )
   {
      iz[i] = iz[i-1] + 1;
      if ( iz[i] < noz )
         nz++;
      else
         break;
   }


   no = nx*ny*nz;
   *index = new int[no];
   *ipf = new double[no];


   itp = 0;
   for ( k=0; k<nz; k++ )
   {
      idz = nox*noy * iz[k];
      ipfz = 1.0;
      pk = iz[k]*delta_z + lpos_z;
      for ( m=0; m<nz; m++ )
      {
         pm = iz[m]*delta_z + lpos_z;
         if ( m != k ) ipfz *= ( z - pm ) / ( pk - pm );
      }

      for ( j=0; j<ny; j++ )
      {
         idy = nox * iy[j];
         ipfy = 1.0;
         pk = iy[j]*delta_y + lpos_y;
         for ( m=0; m<ny; m++ )
         {
            pm = iy[m]*delta_y + lpos_y;
            if ( m != j ) ipfy *= ( y - pm ) / ( pk - pm );
         }

         for ( i=0; i<nx; i++ )
         {
            idx = ix[i];
            ipfx = 1.0;
            pk = ix[i]*delta_x + lpos_x;
            for ( m=0; m<nx; m++ )
            {
               pm = ix[m]*delta_x + lpos_x;
               if ( m != i ) ipfx *= ( x - pm ) / ( pk - pm );
            }

            (*index)[itp] = idx + idy + idz;
            (*ipf)[itp] = ipfx * ipfy * ipfz;

            itp++;
         }
      }
   }

}

void FIELDx::E_inc( double x, double y, double z, complex<double> E[3] )
{
   int no, *index;
   double *ipf;

   itpl( x, y, z, no, &index, &ipf );

   E[0] = E[1] = E[2] = complex<double>(0.0,0.0);
   for ( int i=0; i<no; i++ )
   {
      E[0] += Einc[ index[i] ].x * ipf[i];
      E[1] += Einc[ index[i] ].y * ipf[i];
      E[2] += Einc[ index[i] ].z * ipf[i];
   }

   delete [] index;
   delete [] ipf;
}

void FIELDx::E_scatt( double x, double y, double z, complex<double> E[3] )
{
   int no, *index;
   double *ipf;

   itpl( x, y, z, no, &index, &ipf );

   bool srcf = true;

   E[0] = E[1] = E[2] = complex<double>(0.0,0.0);
   for ( int i=0; i<no; i++ )
   {
      if ( sqrt( abs( J[ index[i] ].x*conj(J[ index[i] ].x)
                      + J[ index[i] ].y*conj(J[ index[i] ].y)
                      + J[ index[i] ].z*conj(J[ index[i] ].z) ) )
           > DBL_EPSILON ) srcf = false;

      E[0] += Escatt[ index[i] ].x * ipf[i];
      E[1] += Escatt[ index[i] ].y * ipf[i];
      E[2] += Escatt[ index[i] ].z * ipf[i];
   }

   if ( ! srcf )
   {
      Xlog << endl;
      Xlog << "WARNING in FIELDx::E_scatt(): probe region not source free!" << endl;
      Xlog << endl;
   }

   delete [] index;
   delete [] ipf;
}

void FIELDx::E_total( double x, double y, double z, complex<double> E[3] )
{
   int no, *index;
   double *ipf;

   itpl( x, y, z, no, &index, &ipf );

   bool srcf = true;

   E[0] = E[1] = E[2] = complex<double>(0.0,0.0);
   for ( int i=0; i<no; i++ )
   {
      if ( sqrt( abs( J[ index[i] ].x*conj(J[ index[i] ].x)
                      + J[ index[i] ].y*conj(J[ index[i] ].y)
                      + J[ index[i] ].z*conj(J[ index[i] ].z) ) )
           > DBL_EPSILON ) srcf = false;

      E[0] += (Einc[ index[i] ].x + Escatt[ index[i] ].x) * ipf[i];
      E[1] += (Einc[ index[i] ].y + Escatt[ index[i] ].y) * ipf[i];
      E[2] += (Einc[ index[i] ].z + Escatt[ index[i] ].z) * ipf[i];
   }

   if ( ! srcf )
   {
      Xlog << endl;
      Xlog << "WARNING in FIELDx::E_total(): probe region not source free!" << endl;
      Xlog << endl;
   }

   delete [] index;
   delete [] ipf;
}


void FIELDx::E_Jtotal_sum( double apx, double apy, double apz, complex<double> E[3] )
{
   int fidx;
   complex<double> G[3][3];
   POINT ap, qp;

   E[0] = E[1] = E[2] = complex<double>(0.0,0.0);
   ap.x = apx;
   ap.y = apy;
   ap.z = apz;

   bool srcf = true;

   for ( int i3=0; i3<noz; i3++ )
   {
      qp.z = i3*delta_z + lpos_z;
      for ( int i2=0; i2<noy; i2++ )
      {
         qp.y = i2*delta_y + lpos_y;
         for ( int i1=0; i1<nox; i1++ )
         {
            qp.x = i1*delta_x + lpos_x;
            fidx = i1 + nox * i2 + nox*noy * i3;

            if ( ( ! grf.EF_green( G, ap, qp ) ) &&
                 ( sqrt( abs( J[ fidx ].x*conj(J[ fidx ].x)
                              + J[ fidx ].y*conj(J[ fidx ].y)
                              + J[ fidx ].z*conj(J[ fidx ].z) ) )
                   > DBL_EPSILON ) ) srcf = false;

            E[0] += G[0][0]*J[fidx].x + G[0][1]*J[fidx].y + G[0][2]*J[fidx].z;
            E[1] += G[1][0]*J[fidx].x + G[1][1]*J[fidx].y + G[1][2]*J[fidx].z;
            E[2] += G[2][0]*J[fidx].x + G[2][1]*J[fidx].y + G[2][2]*J[fidx].z;
         }
      }
   }

   if ( ! srcf )
   {
      Xlog << endl;
      Xlog << "WARNING in FIELDx::E_Jtotal_sum(): probe region not source free!" << endl;
      Xlog << endl;
   }

}
