//
// Class for greens function
// (c) 27.6.2007 Ingo D. Rullhusen
//

#include <float.h>
#include <math.h>
#include <fftw3.h>

#include "natconst.h"
#include "auxiliary.h"

#include "green.h"


static const complex<double> cNULL(0.0,0.0);
static const complex<double> cONE(1.0,0.0);
static const complex<double> cJ(0.0,1.0);



GREEN::GREEN()
{
   FT_grf = NULL;

   FT_grfExx = FT_grfExy = FT_grfExz = NULL;
   FT_grfEyx = FT_grfEyy = FT_grfEyz = NULL;
   FT_grfEzx = FT_grfEzy = FT_grfEzz = NULL;

   clear();
}

GREEN::~GREEN()
{
   clear();
}


bool GREEN::EP_green( complex<double> &grfvp, complex<double> &grfdiv,
                      const POINT ap, const POINT qp )
{
   bool exact = true;
   double r;
   complex<double> expo;

   r = HYPOT( qp.x-ap.x, qp.y-ap.y, qp.z-ap.z );
   if ( r < R )
   {
      r = R;
      exact = false;
   }

   expo = exp( -cJ*k0*r ) / r;

   grfvp = prefac_v * expo;

   grfdiv = prefac_div * expo;

   return(exact);
}

bool GREEN::EF_green( complex<double> G[3][3],
                      const POINT ap, const POINT qp )
{
   bool exact = true;
   double phi, sinphi, cosphi, costh, sinth, xs, ys, zs, r;
   complex<double> Er, Eth, Exs, Eys, Ezs, expo;

// x directed
   xs = ap.y - qp.y;
   ys = ap.z - qp.z;
   zs = ap.x - qp.x;

   phi = atan2( ys, xs ); 
   sinphi = sin(phi);
   cosphi = cos(phi);

   r = sqrt( xs*xs + ys*ys + zs*zs );
   if ( r < R )
   {
      r = R;
      exact = false;
   }

   costh = zs / r;
   sinth = sqrt( 1.0 - costh*costh );

   expo = exp( -cJ * k0 * r );

   Er = costh * 2.0*prefac_E * expo * ( 1.0 + 1.0 / (cJ*k0*r) ) / (r*r);
   Eth = sinth * prefac_E * expo * cJ*k0 * ( 1.0 + 1.0 / (cJ*k0*r) - 1.0 / SQU(k0*r) ) / r;

   Exs = sinth*cosphi * Er + costh*cosphi * Eth;
   Eys = sinth*sinphi * Er + costh*sinphi * Eth;
   Ezs = costh        * Er - sinth        * Eth;

   G[0][0] = Ezs;
   G[1][0] = Exs;
   G[2][0] = Eys;

// y directed
   xs = - ( ap.x - qp.x );
   ys = ap.z - qp.z;
   zs = ap.y - qp.y;

   phi = atan2( ys, xs );
   sinphi = sin(phi);
   cosphi = cos(phi);

//   r = sqrt( xs*xs + ys*ys + zs*zs );
//   if ( r < DBL_EPSILON ) throw( "probe near singularity" );
   costh = zs / r;
   sinth = sqrt( 1.0 - costh*costh );

//   expo = exp( -cJ * k0 * r );

   Er = costh * 2.0*prefac_E * expo * ( 1.0 + 1.0 / (cJ*k0*r) ) / (r*r);
   Eth = sinth * prefac_E * expo * cJ*k0 * ( 1.0 + 1.0 / (cJ*k0*r) - 1.0 / SQU(k0*r) ) / r;

   Exs = sinth*cosphi * Er + costh*cosphi * Eth;
   Eys = sinth*sinphi * Er + costh*sinphi * Eth;
   Ezs = costh        * Er - sinth        * Eth;

   G[0][1] = -Exs;
   G[1][1] =  Ezs;
   G[2][1] =  Eys;

// z directed
   xs = ap.x - qp.x;
   ys = ap.y - qp.y;
   zs = ap.z - qp.z;

   phi = atan2( ys, xs );
   sinphi = sin(phi);
   cosphi = cos(phi);

//   r = sqrt( xs*xs + ys*ys + zs*zs );
//   if ( r < DBL_EPSILON ) throw( "probe near singularity" );
   costh = zs / r;
   sinth = sqrt( 1.0 - costh*costh );

//   expo = exp( -cJ * k0 * r );

   Er = costh * 2.0*prefac_E * expo * ( 1.0 + 1.0 / (cJ*k0*r) ) / (r*r);
   Eth = sinth * prefac_E * expo * cJ*k0 * ( 1.0 + 1.0 / (cJ*k0*r) - 1.0 / SQU(k0*r) ) / r;

   Exs = sinth*cosphi * Er + costh*cosphi * Eth;
   Eys = sinth*sinphi * Er + costh*sinphi * Eth;
   Ezs = costh        * Er - sinth        * Eth;

   G[0][2] = Exs;
   G[1][2] = Eys;
   G[2][2] = Ezs;

   return(exact);
}


void GREEN::FT_EP_green( complex<double> &FTgrfvp, complex<double> &FTgrfdiv,
                         const double kx, const double ky, const double kz )
{
   double kR, k, kq, sinckR, coskR;
   complex<double> expo;

   k = HYPOT( kx, ky, kz );
   kq = k*k;
   kR = k * R;

   if ( kR > DBL_EPSILON ) sinckR = sin(kR) / kR;
   else                    sinckR = 1.0;

   coskR = cos(kR);

   expo = expmjk0R * ( ( sinckR - coskR ) / kq + ( jk0R * sinckR + coskR ) / ( kq - k0q ) );

   FTgrfvp = prefac_v * expo;

   FTgrfdiv = prefac_div * expo;
}


void GREEN::setup( const MainData *MD, double R_ )
{
   int i, j, k;

   k0 = MD->k0;
   R = R_;

   k0q = k0*k0;

   prefac_v = MY0/(4.0*M_PI);
   prefac_div = 1.0/(4.0*M_PI*EPS0);
   prefac_E = ZF0/(4.0*M_PI);
   prefac_ff = cJ*k0*ZF0/(4.0*M_PI);

   prefac_fte = -cJ*MD->omega * prefac_v;

   jk0R = cJ * k0 * R;
   expmjk0R = exp( - jk0R );

   sx = 2 * MD->AIMnox - 1;
   sy = 2 * MD->AIMnoy - 1;
   sz = 2 * MD->AIMnoz - 1;

   delta_k[0] = ( sx > 2 ? 2.0*M_PI / ( sx * MD->delta_x ) : 0.0 );
   delta_k[1] = ( sy > 2 ? 2.0*M_PI / ( sy * MD->delta_y ) : 0.0 );
   delta_k[2] = ( sz > 2 ? 2.0*M_PI / ( sz * MD->delta_z ) : 0.0 );

   delete [] FT_grf;
   FT_grf = new complex<double>[ sx*sy*sz ];

   fftw_plan p;
#pragma omp critical (FFTW)
{
   p = fftw_plan_dft_3d( sz, sy, sx,
                         (fftw_complex*)FT_grf, (fftw_complex*)FT_grf,
                         FFTW_BACKWARD, FFTW_ESTIMATE
                       );
}  // end:pragma omp critical (FFTW)

   int idz, idy, idx;
   double r;
   complex<double> expo;

#pragma omp parallel for private(j,i,r,expo,idz,idy,idx)
   for ( k=0; k<MD->AIMnoz; ++k )
   {
      for ( j=0; j<MD->AIMnoy; ++j )
      {
         for ( i=0; i<MD->AIMnox; ++i )
         {

            r = HYPOT( i*MD->delta_x, j*MD->delta_y, k*MD->delta_z );
            if ( r < R ) r = R;

            expo = exp( -cJ*k0*r ) / r;

            for ( int kq=0; kq< ( k==0 ? 1 : 2 ); kq++ )
            {
            idz = ( kq==0 ? k : sz-k ) * sx * sy;

            for ( int jq=0; jq< ( j==0 ? 1 : 2 ); jq++ )
            {
            idy = ( jq==0 ? j : sy-j ) * sx;

            for ( int iq=0; iq< ( i==0 ? 1 : 2 ); iq++ )
            {
               idx = ( iq==0 ? i : sx-i );

               FT_grf[ idx+idy+idz ] = expo;
            }
            }
            }

         }
      }
   }

   fftw_execute(p);

#pragma omp critical (FFTW)
{
   fftw_destroy_plan(p);
}  // end:pragma omp critical (FFTW)

   delete [] FT_grfExx; FT_grfExx = NULL;
   FT_grfExx = new complex<double>[ sx*sy*sz ];
   delete [] FT_grfExy; FT_grfExy = NULL;
   FT_grfExy = new complex<double>[ sx*sy*sz ];
   delete [] FT_grfExz; FT_grfExz = NULL;
   FT_grfExz = new complex<double>[ sx*sy*sz ];

   delete [] FT_grfEyx; FT_grfEyx = NULL;
   FT_grfEyx = new complex<double>[ sx*sy*sz ];
   delete [] FT_grfEyy; FT_grfEyy = NULL;
   FT_grfEyy = new complex<double>[ sx*sy*sz ];
   delete [] FT_grfEyz; FT_grfEyz = NULL;
   FT_grfEyz = new complex<double>[ sx*sy*sz ];

   delete [] FT_grfEzx; FT_grfEzx = NULL;
   FT_grfEzx = new complex<double>[ sx*sy*sz ];
   delete [] FT_grfEzy; FT_grfEzy = NULL;
   FT_grfEzy = new complex<double>[ sx*sy*sz ];
   delete [] FT_grfEzz; FT_grfEzz = NULL;
   FT_grfEzz = new complex<double>[ sx*sy*sz ];

   fftw_plan pxx, pxy, pxz, pyx, pyy, pyz, pzx, pzy, pzz;
#pragma omp critical (FFTW)
{
   pxx = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfExx, (fftw_complex*)FT_grfExx,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );
   pxy = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfExy, (fftw_complex*)FT_grfExy,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );
   pxz = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfExz, (fftw_complex*)FT_grfExz,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );

   pyx = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfEyx, (fftw_complex*)FT_grfEyx,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );
   pyy = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfEyy, (fftw_complex*)FT_grfEyy,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );
   pyz = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfEyz, (fftw_complex*)FT_grfEyz,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );

   pzx = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfEzx, (fftw_complex*)FT_grfEzx,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );
   pzy = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfEzy, (fftw_complex*)FT_grfEzy,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );
   pzz = fftw_plan_dft_3d( sz, sy, sx,
                           (fftw_complex*)FT_grfEzz, (fftw_complex*)FT_grfEzz,
                           FFTW_BACKWARD, FFTW_ESTIMATE
                         );
}  // end:pragma omp critical (FFTW)

   complex<double> G[3][3];

   POINT ap, qp;
   qp.x = qp.y = qp.z = 0.0;

#pragma omp parallel for private(j,i,ap,G,idx)
   for ( k=0; k<sz; ++k )
   {
      ap.z = ( k > sz/2 ? k-sz : k) * MD->delta_z;

      for ( j=0; j<sy; ++j )
      {
         ap.y = ( j > sy/2 ? j-sy : j ) * MD->delta_y;

         for ( i=0; i<sx; ++i )
         {
            ap.x = ( i > sx/2 ? i-sx : i ) * MD->delta_x;

            EF_green( G, ap, qp );

            idx = i + sx*j + sx*sy*k;

            FT_grfExx[ idx ] = G[0][0];
            FT_grfExy[ idx ] = G[0][1];
            FT_grfExz[ idx ] = G[0][2];

            FT_grfEyx[ idx ] = G[1][0];
            FT_grfEyy[ idx ] = G[1][1];
            FT_grfEyz[ idx ] = G[1][2];

            FT_grfEzx[ idx ] = G[2][0];
            FT_grfEzy[ idx ] = G[2][1];
            FT_grfEzz[ idx ] = G[2][2];
         }
      }
   }

   fftw_execute(pxx);
   fftw_execute(pxy);
   fftw_execute(pxz);

   fftw_execute(pyx);
   fftw_execute(pyy);
   fftw_execute(pyz);

   fftw_execute(pzx);
   fftw_execute(pzy);
   fftw_execute(pzz);

#pragma omp critical (FFTW)
{
   fftw_destroy_plan(pxx);
   fftw_destroy_plan(pxy);
   fftw_destroy_plan(pxz);

   fftw_destroy_plan(pyx);
   fftw_destroy_plan(pyy);
   fftw_destroy_plan(pyz);

   fftw_destroy_plan(pzx);
   fftw_destroy_plan(pzy);
   fftw_destroy_plan(pzz);
}  // end:pragma omp critical (FFTW)

}

void GREEN::clear()
{
   k0 = 0.0;
   R = DBL_MAX;

   k0q = k0*k0;

   prefac_v = MY0/(4.0*M_PI);
   prefac_div = 1.0/(4.0*M_PI*EPS0);
   prefac_E = ZF0/(4.0*M_PI);
   prefac_ff = cJ*k0*ZF0/(4.0*M_PI);

   prefac_fte = 0.0;

   jk0R = cJ * k0 * R;
   expmjk0R = exp( - jk0R );

   delete [] FT_grf;
   FT_grf = NULL;

   delete [] FT_grfExx;
   FT_grfExx = NULL;
   delete [] FT_grfExy;
   FT_grfExy = NULL;
   delete [] FT_grfExz;
   FT_grfExz = NULL;

   delete [] FT_grfEyx;
   FT_grfEyx = NULL;
   delete [] FT_grfEyy;
   FT_grfEyy = NULL;
   delete [] FT_grfEyz;
   FT_grfEyz = NULL;

   delete [] FT_grfEzx;
   FT_grfEzx = NULL;
   delete [] FT_grfEzy;
   FT_grfEzy = NULL;
   delete [] FT_grfEzz;
   FT_grfEzz = NULL;

   sx = sy = sz = 0;
}

void GREEN::FFT_EP_green( complex<double> &FTgrfvp, complex<double> &FTgrfdiv,
                          int &index,
                          const int kx, const int ky, const int kz )
{
   index = kx + sx*ky + sx*sy*kz;

   FTgrfvp = prefac_v * FT_grf[index];
   FTgrfdiv = prefac_div * FT_grf[index];
}


void GREEN::FFT_EF_green( complex<double> G[3][3],
                          int &index,
                          const int kx, const int ky, const int kz )
{
   index = kx + sx*ky + sx*sy*kz;

   G[0][0] = FT_grfExx[index];
   G[0][1] = FT_grfExy[index];
   G[0][2] = FT_grfExz[index];

   G[1][0] = FT_grfEyx[index];
   G[1][1] = FT_grfEyy[index];
   G[1][2] = FT_grfEyz[index];

   G[2][0] = FT_grfEzx[index];
   G[2][1] = FT_grfEzy[index];
   G[2][2] = FT_grfEzz[index];

/*** too inaccurate
   double k[3];

   k[0] = delta_k[0] * ( kx > sx/2 ? kx-sx : kx ) / k0;
   k[1] = delta_k[1] * ( ky > sy/2 ? ky-sy : ky ) / k0;
   k[2] = delta_k[2] * ( kz > sz/2 ? kz-sz : kz ) / k0;

   complex<double> FTgrf;

   FTgrf = prefac_fte * FT_grf[index];

   G[0][0] = FTgrf * ( 1.0 - k[0] * k[0] );
   G[0][1] = FTgrf * (     - k[0] * k[1] );
   G[0][2] = FTgrf * (     - k[0] * k[2] );

   G[1][0] = FTgrf * (     - k[1] * k[0] );
   G[1][1] = FTgrf * ( 1.0 - k[1] * k[1] );
   G[1][2] = FTgrf * (     - k[1] * k[2] );

   G[2][0] = FTgrf * (     - k[2] * k[0] );
   G[2][1] = FTgrf * (     - k[2] * k[1] );
   G[2][2] = FTgrf * ( 1.0 - k[2] * k[2] );
***/
}
