/*
*  Calculation of Legendre polynoms, its zeros, and weights
*  (c) 23.8.2006 Ingo D. Rullhusen
*
* This program is free software; you can redistribute it and/or modify   
* it under the terms of the GNU General Public License as published by   
* the Free Software Foundation; either version 2 of the License, or   
* (at your option) any later version.   
* 
* This program is distributed in the hope that it will be useful,   
* but WITHOUT ANY WARRANTY; without even the implied warranty of   
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   
* GNU General Public License for more details.   
*
* You should have received a copy of the GNU General Public License   
* along with this program; if not, write to the Free Software   
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*
*/

#include <stdlib.h>

#include "zero.h"

#define EPSILON 1e-15


#ifndef NO_SMALL_CAPS

#ifdef NOF77_
#define DGETRF dgetrf
#define DGETRS dgetrs
#else
#define DGETRF dgetrf_
#define DGETRS dgetrs_
#endif

#endif

void DGETRF( int*, int*, double*, int*, int*, int* );
void DGETRS( char*, int*, int*, double*, int*, int*, double*, int*, int* );


double legendre( int L, double x )
{
   int i;
   double Pim1, Pi, Pip1;
   
   Pim1 = 1.0;
   Pi = x;
   Pip1 = 0.0;
   
   if ( L < 0 )
      return(Pip1);
   else if ( L == 0 )
      return(Pim1);
   else if ( L == 1 )
      return(Pi);
   else
   {
      for ( i=1; i<L; i++ )
      {
         Pip1 = (2*i+1) * x * Pi - i * Pim1;
         Pip1 /= i + 1;
         Pim1 = Pi;
         Pi = Pip1;
      }
      return(Pip1);
   }
}

static int legendre_param;
static double tlegendre( double x ){ return(legendre(legendre_param,x)); }

int legendre_zero( double *zeros, int L, double scale )
{
   int i;
   double lb, ub, delta;
   
   legendre_param = L;
   delta = 2.0/L * scale;
   i = 0;
   lb = -1.0;
   ub = delta - 1.0;
   while ( i<L && lb < 1.0 )
   {
      if ( tlegendre(lb)*tlegendre(ub) < 0.0 )
      {
         zeros[i] = zero( tlegendre, lb, ub, EPSILON );
         lb = zeros[i] + delta;
         ub = lb + delta;
         i++;
      }
      else
      {
         lb += delta;
         ub += delta;
      }
   }
   if ( i != L ) return(1);
   return(0);
}

int legendre_weights( double *w, double *x, int L )
{
   char TRANS;         
   int INFO, LDB, *IPIV, LDA, NRHS, N, i, j;
   double *A, *B;

   N = L;
   LDA = N;
   LDB = N;
   NRHS = 1;
   IPIV = calloc( N, sizeof(int) );
   A = calloc( LDA*N, sizeof(double) );
   B = calloc( LDB*NRHS, sizeof(double) );

   for (i=0; i<N; i++)
   {      
      B[i] = 0.0;
      for (j=0; j<N; j++) A[i+N*j] = legendre(i,x[j]);
   }
   B[0] = 2.0;

   TRANS = 'N';
   INFO = 0;
   DGETRF( &N, &N, A, &LDA, IPIV, &INFO );
   if ( INFO )
   {
      return(1);
   }
   DGETRS( &TRANS, &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO );
   if ( INFO )
   {
      return(2);
   }

   for (i=0; i<N; i++) w[i] = B[i];

   free(IPIV);
   free(A);
   free(B);

   return(0);
}
