//
// Class for near part matrix
// (c) 8.6.2007 Ingo D. Rullhusen
//

#include <float.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#include <vector>
#include <iostream>

#include <qprogressdialog.h>

#include "nearmatrix.h"
#include "module_interface.h"
#include "auxiliary.h"
#include "auxiliary_qt.h"

#include "global_vars.h"


static const complex<double> cNULL(0.0,0.0);
static const complex<double> cONE(1.0,0.0); 
static const complex<double> cJ(0.0,1.0);   


NearMatrix::NearMatrix()
{
   Z = NULL;
   L = NULL;
   no_of_modules = 0;
   modno = NULL;
   start_idx_mod = new int[1];
   start_idx_mod[0] = 0;
}

NearMatrix::~NearMatrix()
{
   clear();
   delete [] start_idx_mod;
}


void NearMatrix::setup()
{
   int i, m;
   POINT p1, p2;
   QString str;
   QProgressDialog *PW;

   clear();

   no_of_modules = 0;
   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      if ( ip->use_matrix() ) no_of_modules++;
   }

   modno = new int[ no_of_modules ];
   delete [] start_idx_mod;
   start_idx_mod = new int[ no_of_modules + 1 ];

   m = 0;
   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      if ( ip->use_matrix() )
      {
#ifdef BOUNDARY_CHECK
         if ( m >= no_of_modules )
            throw( "NearMatrix::setup index problem" );
#endif
         modno[m] = ip - used_modules.begin();
         start_idx_mod[m] = ip->matrix_offset();
         m++;
      }
   }
   if ( m == 0 )
      start_idx_mod[m] = 0;
   else
      start_idx_mod[m] = start_idx_mod[m-1]
                         + used_modules[ modno[m-1] ].no_of_bfkt();

// symbolic
   int veno = start_idx_mod[no_of_modules];
   int *eleno, e1, e2, j, k, k1, k2, noele, sno;

   str = QProgressDialog::tr("Calculate matrix symbolically");

   PW = new QProgressDialog( str,
                             QProgressDialog::tr("Cancel"),
                             4*veno-1,
                             ptrMW,
                             QProgressDialog::tr("progress"),
                             true );
   PW->setProgress(0);

   eleno = new int[veno];
   for ( int e=0; e<veno; ++e ) eleno[e] = 0;

   double *cp_x, *cp_y, *cp_z, *s_x, *s_y, *s_z, D;

   for ( k=0; k<no_of_modules; k++ )
   {
      ExternalModule *ip = &used_modules[ modno[k] ];

      noele = ip->no_of_bfkt();

      double *CPx = new double[ noele ];
      double *CPy = new double[ noele ];
      double *CPz = new double[ noele ];
      double *R   = new double[ noele ];

#pragma omp parallel for private(e1,cp_x,cp_y,cp_z,s_x,s_y,s_z,sno)
      for ( i=0; i<noele; ++i )
      {
         double CP[3], r;

         e1 = i + start_idx_mod[k];
#pragma omp critical (PW_nearmatrix_1)
{
         PW->setProgress(e1);
         if ( PW->wasCanceled() )
         {
            delete PW;

            delete [] CPz;
            delete [] CPy;
            delete [] CPx;
            delete [] R;

            Xlog << "MODULE( " << m << ": "
                 << ip->name() << " ): ";
            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
}

         ip->matrix_bfkt_bounding_box( &MD, i, &sno,
                                       &cp_x, &cp_y, &cp_z,
                                       &s_x, &s_y, &s_z );

         umkugel( sno, cp_x, cp_y, cp_z, s_x, s_y, s_z, r, CP );

         CPx[i] = CP[0];
         CPy[i] = CP[1];
         CPz[i] = CP[2];
         R[i]   = r;

         delete [] cp_x;
         delete [] cp_y;
         delete [] cp_z;
         delete [] s_x;
         delete [] s_y;
         delete [] s_z;
      }

#pragma omp parallel for private(e1, D,j)
      for ( i=0; i<noele; ++i )
      {
         e1 = i + start_idx_mod[k];
#pragma omp critical (PW_nearmatrix_2)
{
         PW->setProgress( veno+e1 );
         if ( PW->wasCanceled() )
         {
            delete PW;

            delete [] CPz;
            delete [] CPy;
            delete [] CPx;
            delete [] R;

            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
}

         D = MD.wavelength*MD.near_limit + R[i];

         for ( j=i; j<noele; ++j )
         {
            if ( HYPOT( CPx[j]-CPx[i], CPy[j]-CPy[i], CPz[j]-CPz[i] ) < D + R[j] )
            {
               eleno[ e1 ]++;
            }
         }

      }

      delete [] CPz;
      delete [] CPy;
      delete [] CPx;
      delete [] R;
   }

   noele = 0;
#pragma omp parallel for reduction(+:noele)
   for ( e1=0; e1<veno; e1++ ) noele += eleno[e1];

   delete Z;
   Z = new SparseMatrix< complex<double> >( true, veno, noele );

   Z->first_ele[0] = 0;
   for ( int e=1; e<=veno; e++ )
      Z->first_ele[e] = Z->first_ele[e-1] + eleno[e-1];

   delete [] eleno;

   for ( k=0; k<no_of_modules; k++ )
   {
      ExternalModule *ip = &used_modules[ modno[k] ];

      noele = ip->no_of_bfkt();

      double *CPx = new double[ noele ];
      double *CPy = new double[ noele ];
      double *CPz = new double[ noele ];
      double *R   = new double[ noele ];

#pragma omp parallel for private(e1,cp_x,cp_y,cp_z,s_x,s_y,s_z,sno)
      for ( i=0; i<noele; ++i )
      {
         double CP[3], r;

         e1 = i + start_idx_mod[k];
#pragma omp critical (PW_nearmatrix_3)
{
         PW->setProgress( 2*veno+e1 );
         if ( PW->wasCanceled() )
         {
            delete PW;

            delete [] CPz;
            delete [] CPy;
            delete [] CPx;
            delete [] R;

            Xlog << "MODULE( " << m << ": "
                 << ip->name() << " ): ";
            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
}

         ip->matrix_bfkt_bounding_box( &MD, i, &sno,
                                       &cp_x, &cp_y, &cp_z,
                                       &s_x, &s_y, &s_z );

         umkugel( sno, cp_x, cp_y, cp_z, s_x, s_y, s_z, r, CP );

         CPx[i] = CP[0];
         CPy[i] = CP[1];
         CPz[i] = CP[2];
         R[i]   = r;

         delete [] cp_x;
         delete [] cp_y;
         delete [] cp_z;
         delete [] s_x;
         delete [] s_y;
         delete [] s_z;
      }

#pragma omp parallel for private(e1,e2,k1,m, D,j)
      for ( i=0; i<noele; ++i )
      {
         e1 = i + start_idx_mod[k];
#pragma omp critical (PW_nearmatrix_4)
{
         PW->setProgress( 3*veno+e1 );
         if ( PW->wasCanceled() )
         {
            delete PW;

            delete [] CPz;
            delete [] CPy;
            delete [] CPx;
            delete [] R;

            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
}

         k1 = 0;

         D = MD.wavelength*MD.near_limit + R[i];

         for ( j=i; j<noele; ++j )
         {
            e2 = j + start_idx_mod[k];

            if ( HYPOT( CPx[j]-CPx[i], CPy[j]-CPy[i], CPz[j]-CPz[i] ) < D + R[j] )
            {

               m = Z->first_ele[e1] + k1;
#ifdef BOUNDARY_CHECK
               if( m >= Z->first_ele[e1+1] )
                  throw( "NearMatrix::setup index problem" );
#endif
               Z->colid[m] = e2;

               ++k1;
            }
         }

      }

      delete [] CPz;
      delete [] CPy;
      delete [] CPx;
      delete [] R;
   }

   PW->cancel();
   delete PW;


// numeric
   for ( k=0; k<no_of_modules; k++ )
   {
      ExternalModule *ip = &used_modules[ modno[k] ];

      try
      {
         if ( ! thread_of_ele_setup(&*ip) )
            throw( "canceled" );
      }
      catch ( const char *error )
      {
         Xlog << "MODULE( " << modno[k]+1 << ": "
              << ip->name() << " ): ";
         throw(error);
      }

      str = QProgressDialog::tr("Calculate numerically module no. %1").arg(modno[k]+1);

      PW = new QProgressDialog( str,
                                QProgressDialog::tr("Cancel"),
                                ip->no_of_bfkt()-1,
                                ptrMW,
                                QProgressDialog::tr("progress"),
                                true );
      PW->setProgress(0);

      m = ip->no_of_bfkt();
#pragma omp parallel for private(e1,j)
      for ( i=0; i<m; i++ )
      {
         e1 = i + start_idx_mod[k];
#pragma omp critical (PW_nm_1)
{
         PW->setProgress( i );
         if ( PW->wasCanceled() )
         {
            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
}

         for ( j=Z->first_ele[e1]; j<Z->first_ele[e1+1]; j++ )
         {
            Z->ele[j] = ip->matrix_ele( &MD,
                                        i,
                                        Z->colid[j] - start_idx_mod[k] );
         }

      }

      PW->cancel();
      delete PW;

      ip->matrix_ele_cleanup( &MD );
   }


// precon
   complex<double> lii, csum, aij;

   if ( MD.precon > 0 )  // Cholesky
   {
      unsigned int ui = 0;
      str = QProgressDialog::tr("Calculate preconditioner");
      PW = new QProgressDialog( str,
                                QProgressDialog::tr("Cancel"),
                                4*veno-1,
                                ptrMW,
                                QProgressDialog::tr("progress"),
                                true );

// allocate
      k = 0;
#pragma omp parallel for reduction(+:k)
      for ( e1=0; e1<veno; e1++ )
      {
#pragma omp critical (PW_nm_2)
{
         PW->setProgress(ui);
         if ( PW->wasCanceled() )
         {
            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
         ui++;
}

         if ( e1 < MD.precon )
            k += e1+1;
         else
            k += MD.precon;
      }
      delete L;
      L = new SparseMatrix< complex<float> >( false, veno, k );

      k = 0;
      for ( e1=0; e1<veno; e1++ )
      {
         PW->setProgress(ui);
         if ( PW->wasCanceled() )
         {
            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
         ui++;

         L->first_ele[e1] = k;
         if ( e1 < MD.precon )
            k += e1 + 1;
         else
            k += MD.precon;
      }
      L->first_ele[veno] = k;

#pragma omp parallel for
      for ( e1=0; e1<veno; e1++ )
      {
#pragma omp critical (PW_nm_3)
{
         PW->setProgress(ui);
         if ( PW->wasCanceled() )
         {
            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
         ui++;
}

         for ( int e=L->first_ele[e1]; e<L->first_ele[e1+1]; e++ )
         {
            L->colid[e] = e1 - ( L->first_ele[e1+1]-1 - e );
         }
      }

// calculate
///#pragma omp parallel private(noele,csum,i,k,k1,k2,aij) trouble with throw
///{
      for ( i=0; i<veno; i++ )
      {
         PW->setProgress(ui);
         if ( PW->wasCanceled() )
         {
            clear();
            throw( "matrix calculation canceled" );
         }
         ptrAPP->processEvents();
         ui++;

// L_ii
         csum = cNULL;
///#pragma omp single
///{
         for ( int e=L->first_ele[i]; e<L->first_ele[i+1]-1; e++ )
         {
            csum += L->ele[e]*L->ele[e];
         }

         lii = sqrt( Z->ele[ Z->first_ele[i] ] - csum );
         if ( abs(lii) < DBL_EPSILON )
         {
            PW->cancel();
            throw( "Z not positive definite" );
         }
         L->ele[ L->first_ele[i+1]-1 ] = lii;
///}   // end pragma omp single

// L_ji
         noele = ( veno<i+MD.precon ? veno : i+MD.precon );
///#pragma omp for
         for ( j=i+1; j<noele; j++ )
         {

            csum = cNULL;
            k1 = L->first_ele[i];
            k2 = L->first_ele[j];
            while ( L->colid[k1] < L->colid[k2] && L->colid[k1] < i )
            {
               k1++;
            }
#ifdef BOUNDARY_CHECK
            if ( L->colid[k1] != L->colid[k2] && L->colid[k1] < i )
            {
               cerr << "Error in NearMatrix::setup(): line "
                          << __LINE__ << "!" << endl;
               exit(1);
            }
#endif
            while ( L->colid[k1] < i )
            {
               csum += L->ele[k1]*L->ele[k2];
               k1++; k2++;
            }

            for ( k=Z->first_ele[i]; k<Z->first_ele[i+1]; k++ )
            {
               if ( Z->colid[k] == j ) break;
            }
            if ( k < Z->first_ele[i+1] )
               aij = Z->ele[k];
            else
               aij = cNULL;


            while ( L->colid[k2] < i ) k2++;
            L->ele[k2] = ( aij - csum ) / lii;

         }

      }
///}   // end pragma omp parallel private(noele,csum,i,k,k1,k2,aij)

      PW->cancel();
      delete PW;

   }
   else  // no precon
   {

      L = new SparseMatrix< complex<float> >( false, veno, veno);
      for ( int e=0; e<veno; e++ )
      {
         L->first_ele[e] = e;
         L->colid[e] = e;
         L->ele[e] = cONE;
      }
      L->first_ele[veno] = veno;

   }


// AIM correction
   str = QProgressDialog::tr("AIM correction");
   PW = new QProgressDialog( str,
                             QProgressDialog::tr("Cancel"),
                             veno-1,
                             ptrMW,
                             QProgressDialog::tr("progress"),
                             true );
   PW->setProgress(0);

   if ( MD.AIMcm == ACM_SPCDOM )
   {

#pragma omp parallel private(k,e1,e2,p1,p2)
{
   int kta, jta, ita, ktq, jtq, itq, idx;
   complex<double> Ax, Ay, Az, Phi, grfv, grfdiv, colel;

#pragma omp for
   for ( e1=0; e1<veno; e1++ )
   {
#pragma omp critical (PW_nm_4a)
{
      PW->setProgress(e1);
}
#ifndef TERMINAL
      if ( PW->wasCanceled() )
      {
         clear();
         throw( "matrix calculation canceled" );
      }
      ptrAPP->processEvents();
#endif

      for ( k=Z->first_ele[e1]; k<Z->first_ele[e1+1]; k++ )
      {
         e2 = Z->colid[k];

         colel = cNULL;
         for ( int ka=0; ka<AD[e1].noz; ka++ )
         {
         kta = ka + AD[e1].z_off;
         p1.z = kta * MD.delta_z + (MD.cpos_z - 0.5*MD.size_z);

         for ( int ja=0; ja<AD[e1].noy; ja++ )
         {
         jta = ja + AD[e1].y_off;
         p1.y = jta * MD.delta_y + (MD.cpos_y - 0.5*MD.size_y);

         for ( int ia=0; ia<AD[e1].nox; ia++ )
         {
         ita = ia + AD[e1].x_off;
         p1.x = ita * MD.delta_x + (MD.cpos_x - 0.5*MD.size_x);

            Ax = cNULL;
            Ay = cNULL;
            Az = cNULL;
            Phi = cNULL;
            for ( int kq=0; kq<AD[e2].noz; kq++ )
            {
            ktq = kq + AD[e2].z_off;
            p2.z = ktq * MD.delta_z + (MD.cpos_z - 0.5*MD.size_z);

            for ( int jq=0; jq<AD[e2].noy; jq++ )
            {
            jtq = jq + AD[e2].y_off;
            p2.y = jtq * MD.delta_y + (MD.cpos_y - 0.5*MD.size_y);

            for ( int iq=0; iq<AD[e2].nox; iq++ )
            {
            itq = iq + AD[e2].x_off;
            p2.x = itq * MD.delta_x + (MD.cpos_x - 0.5*MD.size_x);

               grf.EP_green( grfv, grfdiv, p1, p2 );

               idx = AD[e2].index( iq, jq, kq );
               Ax += grfv*(double)AD[e2].lambda_x[idx];
               Ay += grfv*(double)AD[e2].lambda_y[idx];
               Az += grfv*(double)AD[e2].lambda_z[idx];

               Phi += grfdiv * (double)AD[e2].lambda_div[idx];
            }
            }
            }

            idx = AD[e1].index( ia, ja, ka );

            colel -= cJ*MD.omega * (
                       (double)AD[e1].lambda_x[idx] * Ax
                     + (double)AD[e1].lambda_y[idx] * Ay
                     + (double)AD[e1].lambda_z[idx] * Az
                                   )
                     + (double)AD[e1].lambda_div[idx]
                       * Phi / (cJ*MD.omega);

         }
         }
         }

// subtract
         Z->ele[k] -= colel;
      }

   }

}  // end:pragma omp parallel private(k,e1,e2,p1,p2)

   }
   else if ( MD.AIMcm == ACM_FFT )
   {

#pragma omp parallel private(k,e1,e2)
{
   int idx, k1, k2, k3;
   complex<double> vp, phi;
   AIMfield AF_t( 2*MD.AIMnox-1, 2*MD.AIMnoy-1, 2*MD.AIMnoz-1 );

#pragma omp for
   for ( e1=0; e1<veno; ++e1 )   // using symmetry
   {
#pragma omp critical (PW_nm_4b)
{
      PW->setProgress(e1);
}
#ifndef TERMINAL
      if ( PW->wasCanceled() )
      {
         clear();
         throw( "matrix calculation canceled" );
      }
      ptrAPP->processEvents();
#endif

      AF_t.set2zero();
      AF_t.add( cONE, AD[e1] );
      AF_t.FFT();

      for ( k3=0; k3<AF_t.noz; ++k3 )
      {
         for ( k2=0; k2<AF_t.noy; ++k2 )
         {
            for ( k1=0; k1<AF_t.nox; ++k1 )
            {
               grf.FFT_EP_green( vp, phi, idx, k1, k2, k3 );
               AF_t.X[idx]  *= vp;
               AF_t.Y[idx]  *= vp;
               AF_t.Z[idx]  *= vp;
               AF_t.div[idx] *= phi;
            }
         }
      }

      AF_t.IFFT();

      for ( k=Z->first_ele[e1]; k<Z->first_ele[e1+1]; ++k )
      {
         e2 = Z->colid[k];

         AF_t.mult_lambda( &vp, &phi, &(AD[e2]) );

// subtract - - = +
         Z->ele[k] += cJ*MD.omega * vp + phi / (cJ*MD.omega);
      }

   }

}  // end:pragma omp parallel private(k,e1,e2)

   }
   else
   {
      throw( "unknown AIM correction mode" );
   }


   PW->cancel();
   delete PW;
}


void NearMatrix::clear()
{
   delete Z;
   Z = NULL;
   delete L;
   L = NULL;

   delete [] modno;
   modno = NULL;
   delete [] start_idx_mod;
   start_idx_mod = new int[1];
   start_idx_mod[0] = 0;

   no_of_modules = 0;
}


void NearMatrix::read_binary( fstream &file )
{
   delete Z;
   delete L;

   delete [] modno;
   delete [] start_idx_mod;


   file.read( (char*) this, sizeof(NearMatrix) );

   Z = new SparseMatrix< complex<double> >( file );
   L = new SparseMatrix< complex<float> >( file );


   if ( no_of_modules > 0 )
   {
      modno = new int[no_of_modules];
      file.read( (char*) modno, sizeof(*modno)*no_of_modules );
   }
   else
   {
      modno = NULL;
   }


   start_idx_mod = new int[no_of_modules+1];
   file.read( (char*) start_idx_mod, sizeof(*start_idx_mod)*(no_of_modules+1) );
}
 
void NearMatrix::write_binary( fstream &file ) const
{
   file.write( (char*) this, sizeof(*this) );

   Z->to_binary( file );
   L->to_binary( file );

   file.write( (char*) modno, sizeof(*modno)*no_of_modules );
   file.write( (char*) start_idx_mod, sizeof(*start_idx_mod)*(no_of_modules+1) );
}
