//
// Base class for OpenGL drawing
// (c) 23.7.2007 Ingo D. Rullhusen
//

#include "ogldrawing.h"

#define DEPTH_Z -4.5
#define NEAR_Z   3.0
#define FAR_Z   30.0
#define SCREEN_Z 4.5
#define EYESEP2  0.2
#define BOXSIZE  1.5




OGLdrawing::OGLdrawing( QWidget* parent, const char* name )
: QGLWidget( QGLFormat( AccumBuffer ), parent, name )
{
   HRot = VRot = 0.0;

   initHaxis[0] = 1.0;
   initHaxis[1] = 0.0;
   initHaxis[2] = 0.0;

   initVaxis[0] = 0.0;
   initVaxis[1] = 1.0;
   initVaxis[2] = 0.0;

   anaglyph = false;
   bkgC = Qt::black;

   R0 = R = 1.0;

   scale_x = scale_y = scale_z = 1.0;

   pixels = NULL;

   object = 0;
}

OGLdrawing::~OGLdrawing()
{
   makeCurrent();
   glDeleteLists( object, 1 );
   doneCurrent();

   delete [] (char*)pixels;
}



void OGLdrawing::paintGL()
{
// Fortran scheme
   const GLfloat lclr[] = { 0.5,  0.0, 0.0, 0.0,
                            0.25, 0.0, 0.0, 0.0,
                            0.25, 0.0, 0.0, 0.0,
                            0.0,  0.0, 0.0, 1.0 };

   const GLfloat rclr[] = { 0.0, 0.4, 0.4, 0.0,
                            0.0, 0.6, 0.0, 0.0,
                            0.0, 0.0, 0.6, 0.0,
                            0.0, 0.0, 0.0, 1.0 };

   glMatrixMode( GL_MODELVIEW );
   glLoadMatrixf( TM );
   glMultMatrixf( RM );
   glRotatef( HRot, initHaxis[0], initHaxis[1], initHaxis[2] );
   glRotatef( VRot, initVaxis[0], initVaxis[1], initVaxis[2] );

   glMatrixMode(GL_PROJECTION);
   if ( anaglyph )
   {
// left
      glClearColor(0.0,0.0,0.0,0.0);
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

      glLoadMatrixf(PML);
      glCallList( object );
   }
   else
   {
// mono
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

      glLoadMatrixf(PMC);
      glCallList( object );
      return;
   }

   glReadPixels(0,0,w,h,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
   glMatrixMode(GL_COLOR);
   glLoadMatrixf(lclr);
   glDrawPixels(w,h,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
   glLoadIdentity();

   glAccum(GL_LOAD,1.0);

// right
   qglClearColor( bkgC );
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   glMatrixMode(GL_PROJECTION);
   glLoadMatrixf(PMR);
   glCallList( object );

   glReadPixels(0,0,w,h,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
   glMatrixMode(GL_COLOR);
   glLoadMatrixf(rclr);
   glDrawPixels(w,h,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
   glLoadIdentity();

// mix
   glAccum(GL_ACCUM,0.9);
   glAccum(GL_RETURN,1.0);
}

void OGLdrawing::reset_rotation()
{
   HRot = VRot = 0.0;
   R = R0;

   update_view_radius();

   glMatrixMode( GL_MODELVIEW );

   glLoadIdentity();
   glGetFloatv( GL_MODELVIEW_MATRIX, RM );

   updateGL();
}

void OGLdrawing::initializeGL()
{
   const GLfloat amb[4]  = {0.6, 0.6, 0.6, 1.0 };

   qglClearColor( bkgC );

   glEnable( GL_DEPTH_TEST );
   glEnable( GL_LIGHTING );
   glEnable( GL_LIGHT0 );
   glEnable( GL_NORMALIZE );

   glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, 1 );
   glLightfv( GL_LIGHT0, GL_AMBIENT, amb );

   glMatrixMode( GL_MODELVIEW );

   glLoadIdentity();
   glGetFloatv( GL_MODELVIEW_MATRIX, RM );
}

void OGLdrawing::resizeGL( int w_, int h_ )
{
   const GLfloat lpos[4] = {-1.0, 1.0, 1.0, 0.0 };

   w = w_;
   h = h_;
   delete [] (char*)pixels;
   pixels = (GLvoid*) new char[ sizeof(GLubyte) * 4 * w * h ];

   glViewport( 0, 0, (GLint)w, (GLint)h );

   scale_x = ( w > h ? (GLfloat)h/(GLfloat)w : 1.0 );
   scale_y = ( w < h ? (GLfloat)w/(GLfloat)h : 1.0 );
   scale_z = 1.0;

   glLightfv( GL_LIGHT0, GL_POSITION, lpos );

   update_view_radius();

   updateGL();
}


void OGLdrawing::setViewRadius( double r )
{
   R = R0 = r;
   update_view_radius();
}

double OGLdrawing::ViewRadius() const
{
   return(R);
}

void OGLdrawing::update_view_radius()
{
   glMatrixMode( GL_PROJECTION );

// mono
   glLoadIdentity();
   glFrustum(-0.5*BOXSIZE*R, 0.5*BOXSIZE*R, -0.5*BOXSIZE*R, 0.5*BOXSIZE*R, NEAR_Z*R, FAR_Z*R);
   glGetFloatv( GL_PROJECTION_MATRIX, PMC );

   double frustumshift = EYESEP2*NEAR_Z/SCREEN_Z;
// left
   glLoadIdentity();
   glFrustum( (-0.5*BOXSIZE+frustumshift)*R, (0.5*BOXSIZE+frustumshift)*R,
              -0.5*BOXSIZE*R, 0.5*BOXSIZE*R, NEAR_Z*R, FAR_Z*R );
   glTranslatef( +EYESEP2*R, 0.0, 0.0);
   glGetFloatv( GL_PROJECTION_MATRIX, PML );

// right
   glLoadIdentity();
   glFrustum( (-0.5*BOXSIZE-frustumshift)*R, (0.5*BOXSIZE-frustumshift)*R,
              -0.5*BOXSIZE*R, 0.5*BOXSIZE*R, NEAR_Z*R, FAR_Z*R );
   glTranslatef( -EYESEP2*R, 0.0, 0.0);
   glGetFloatv( GL_PROJECTION_MATRIX, PMR );


   glMatrixMode( GL_MODELVIEW );

   glLoadIdentity();
   glScalef( scale_x, scale_y, scale_z );
   glTranslatef( 0.0, 0.0, DEPTH_Z*R );
   glGetFloatv( GL_MODELVIEW_MATRIX, TM );
}


void OGLdrawing::setBkgColor( QColor c )
{
   bkgC = c;
   qglClearColor( bkgC );
   updateGL();
}

QColor OGLdrawing::BkgColor() const
{
   return(bkgC);
}

void OGLdrawing::setAnaglyph( bool a )
{
   anaglyph = a;
   updateGL();
}

bool OGLdrawing::Anaglyph() const
{
   return(anaglyph);
}


void OGLdrawing::focusInEvent( QFocusEvent* e )
{
   if ( e->gotFocus() ) emit gotFocus();
}


void OGLdrawing::mousePressEvent ( QMouseEvent *e )
{
   initMousePos = e->pos();

   if ( e->button() == Qt::RightButton )
   {
      mpy = initMousePos.y();
   }
   else if ( e->button() == Qt::LeftButton )
   {
      glMatrixMode( GL_MODELVIEW );

      glLoadMatrixf( RM );
      glRotatef( HRot, initHaxis[0], initHaxis[1], initHaxis[2] );
      glRotatef( VRot, initVaxis[0], initVaxis[1], initVaxis[2] );

      glGetFloatv( GL_MODELVIEW_MATRIX, RM );
   
      initHaxis[0] = RM[0];
      initHaxis[1] = RM[4];
      initHaxis[2] = RM[8];
      initVaxis[0] = RM[1];
      initVaxis[1] = RM[5];
      initVaxis[2] = RM[9];

      HRot = VRot = 0.0;
   }
   else
   {
      e->ignore();
      return;
   }

   e->accept();
}

void OGLdrawing::mouseMoveEvent ( QMouseEvent *e )
{
   QPoint MP = e->pos();

   if ( e->state() == Qt::RightButton )
   {
      GLfloat d = (GLfloat)( MP.y() - mpy );

      if      ( d > 0 ) R /= 1.0 + 0.01*d;
      else if ( d < 0 ) R *= 1.0 - 0.01*d;

      update_view_radius();

      mpy = MP.y();
   }
   else if ( e->state() == Qt::LeftButton )
   {
      VRot = (GLfloat)( (MP.x()-initMousePos.x()) );
      HRot = (GLfloat)( (MP.y()-initMousePos.y()) );
   }
   else
   {
      e->ignore();
      return;
   }

   e->accept();

   updateGL();
}

void OGLdrawing::wheelEvent( QWheelEvent *e )
{
   if      ( e->delta() > 0 ) R *= 1.1;
   else if ( e->delta() < 0 ) R /= 1.1;
   else                       return;

   update_view_radius();

   e->accept();

   updateGL();
}
