//
// OpenGL Widget
// (c) 10.11.2007 Ingo D. Rullhusen
//

#include <vector>

#include <qimage.h>
#include <qfiledialog.h>

#include "global_vars.h"

#include "oglw.h"


#define MAX_OGL_OBJ 4


OGLW::OGLW( QWidget* parent, const char* name )
: OGLdrawing::OGLdrawing( parent, name )
{
   makeCurrent();
   partlist = glGenLists( MAX_OGL_OBJ );
}

OGLW::~OGLW()
{
   makeCurrent();
   glDeleteLists( partlist, MAX_OGL_OBJ );
}


void OGLW::clearall()
{
   makeCurrent();
   glDeleteLists( object, 1 );
   object = glGenLists( 1 );
}


void OGLW::setup_geo()
{
   double R;
   GLuint gl_list;

   R = 0.5*sqrt( MD.size_x*MD.size_x
                 + MD.size_y*MD.size_y
                 + MD.size_z*MD.size_z );

   makeCurrent();

// matrix
   gl_list = partlist + 0;
   glNewList( gl_list, GL_COMPILE );

   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      if ( ip->use_matrix() )
      {
         try { ip->matrix_add_to_OGLlist( &MD, gl_list ); }
         catch ( const char *error )
         {
            Xlog << "MODULE( " << (ip-used_modules.begin())+1 << ": "
                 << ip->name() << " ): "
                 << error << "!" << endl;
            break;
         }
      }
   }

   glEndList();

// RHS
   gl_list = partlist + 1;
   glNewList( gl_list, GL_COMPILE );

   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      if ( ip->use_RHS() )
      {
         try { ip->RHS_add_to_OGLlist( &MD, gl_list ); }
         catch ( const char *error )
         {
            Xlog << "MODULE( " << (ip-used_modules.begin())+1 << ": "
                 << ip->name() << " ): "
                 << error << "!" << endl;
            break;
         }
      }
   }

   glEndList();

// extract
   gl_list = partlist + 2;
   glNewList( gl_list, GL_COMPILE );

   for ( vector<ExternalModule>::iterator ip = used_modules.begin();
         ip != used_modules.end();
         ip++ )
   {
      if ( ip->use_extract() )
      {
         try { ip->extract_add_to_OGLlist( &MD, gl_list ); }
         catch ( const char *error )
         {
            Xlog << "MODULE( " << (ip-used_modules.begin())+1 << ": "
                 << ip->name() << " ): "
                 << error << "!" << endl;
            break;
         }
      }
   }


   if ( MD.near_field )
   {
      const GLfloat clr[4]   = { 1.00, 0.00, 1.00, 1.0 };

      glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr );
      
      glBegin(GL_LINES);

      glVertex3d( MD.nfp_x - 0.1*R, MD.nfp_y, MD.nfp_z );
      glVertex3d( MD.nfp_x + 0.1*R, MD.nfp_y, MD.nfp_z );

      glVertex3d( MD.nfp_x, MD.nfp_y - 0.1*R, MD.nfp_z );
      glVertex3d( MD.nfp_x, MD.nfp_y + 0.1*R, MD.nfp_z );

      glVertex3d( MD.nfp_x, MD.nfp_y, MD.nfp_z - 0.1*R );
      glVertex3d( MD.nfp_x, MD.nfp_y, MD.nfp_z + 0.1*R );

      glEnd();
   }

   glEndList();

// AIM grid
   gl_list = partlist + 3;
   glNewList( gl_list, GL_COMPILE );

   double xp, yp, zp, d;
   const GLfloat clrg[4] = { 1.00, 0.00, 1.00, 1.0 };

   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clrg );
   glBegin(GL_LINES);

   d = 0.1 * (MD.size_x + MD.size_y + MD.size_z)
             / ( MD.AIMnox + MD.AIMnoy + MD.AIMnoz);

   for ( int k=0; k<MD.AIMnoz; k++ )
   {
      zp = MD.cpos_z - 0.5*MD.size_z + k*MD.delta_z;

      for ( int j=0; j<MD.AIMnoy; j++ )
      {
         yp = MD.cpos_y - 0.5*MD.size_y + j*MD.delta_y;

         for ( int i=0; i<MD.AIMnox; i++ )
         {
            xp = MD.cpos_x - 0.5*MD.size_x + i*MD.delta_x;

            glVertex3d( xp - d, yp, zp );
            glVertex3d( xp + d, yp, zp );

            glVertex3d( xp, yp - d, zp );
            glVertex3d( xp, yp + d, zp );

            glVertex3d( xp, yp, zp - d );
            glVertex3d( xp, yp, zp + d );
         }
      }
   }

   glEnd();

   glEndList();


// circumsphere
   setViewRadius(R);

}


void OGLW::display_geo( bool mat, bool rhs, bool extract, bool grid )
{
   clearall();

   glNewList( object, GL_COMPILE );

   if ( mat )     glCallList( partlist + 0 );
   if ( rhs )     glCallList( partlist + 1 );
   if ( extract ) glCallList( partlist + 2 );
   if ( grid )    glCallList( partlist + 3 );

   glEndList();

   updateGL();
}




QString OGLW::GetSavePicName()
{
   QString fn;
   QFileDialog fd( MD.PrjPath,
                   tr("All files (*)"),
                   this,
                   tr("save image dialog"),
                   true );
   fd.setMode( QFileDialog::AnyFile );
   fd.setCaption(tr("Save image"));

   QStringList list = QImage::outputFormatList();
   QStringList::Iterator it = list.begin();
   while( it != list.end() )
   {
      fd.addFilter( tr("Image (*.") + *it + " *." + (*it).lower() + ")" );
      ++it;
   }
   fd.setSelectedFilter(tr("All files (*)"));

   if ( fd.exec() == QDialog::Accepted )
      fn = fd.selectedFile();
   else
      fn = "";

   return(fn);
}



void OGLW::export_image()
{
   QImage IMG = grabFrameBuffer(true);
   QString fn = GetSavePicName();

   if ( fn != "" )
   {
      QString ext = fn.section( '.', -1 );

      if ( IMG.save( fn, ext.upper(), 100 ) )
      {
         Xlog << "image exported as \"" << fn << "\"." << endl;
      }
      else
      {
         Xlog << "cannot save \"" << fn << "\", wrong extension?" << endl;
      }
   }

}
